@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2015-2023 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -33,6 +33,8 @@
 #include <QJsonArray>
 #include "qzmqsocket.h"
 #include "qzmqvalve.h"
+#include "qzmqreqmessage.h"
+#include "qtcompat.h"
 #include "tnetstring.h"
 #include "rtimer.h"
 #include "log.h"
@@ -138,7 +140,7 @@ public:
 		{
 			QVariantHash args = req->args();
 
-			if(!args.contains("method") || args["method"].type() != QVariant::ByteArray)
+			if(!args.contains("method") || typeId(args["method"]) != QMetaType::QByteArray)
 			{
 				respondError("bad-request");
 				return;
@@ -146,7 +148,7 @@ public:
 
 			requestData.method = QString::fromLatin1(args["method"].toByteArray());
 
-			if(!args.contains("uri") || args["uri"].type() != QVariant::ByteArray)
+			if(!args.contains("uri") || typeId(args["uri"]) != QMetaType::QByteArray)
 			{
 				respondError("bad-request");
 				return;
@@ -159,7 +161,7 @@ public:
 				return;
 			}
 
-			if(!args.contains("headers") || args["headers"].type() != QVariant::List)
+			if(!args.contains("headers") || typeId(args["headers"]) != QMetaType::QVariantList)
 			{
 				respondError("bad-request");
 				return;
@@ -167,14 +169,14 @@ public:
 
 			foreach(const QVariant &vheader, args["headers"].toList())
 			{
-				if(vheader.type() != QVariant::List)
+				if(typeId(vheader) != QMetaType::QVariantList)
 				{
 					respondError("bad-request");
 					return;
 				}
 
 				QVariantList vlist = vheader.toList();
-				if(vlist.count() != 2 || vlist[0].type() != QVariant::ByteArray || vlist[1].type() != QVariant::ByteArray)
+				if(vlist.count() != 2 || typeId(vlist[0]) != QMetaType::QByteArray || typeId(vlist[1]) != QMetaType::QByteArray)
 				{
 					respondError("bad-request");
 					return;
@@ -183,7 +185,7 @@ public:
 				requestData.headers += HttpHeader(vlist[0].toByteArray(), vlist[1].toByteArray());
 			}
 
-			if(!args.contains("body") || args["body"].type() != QVariant::ByteArray)
+			if(!args.contains("body") || typeId(args["body"]) != QMetaType::QByteArray)
 			{
 				respondError("bad-request");
 				return;
@@ -194,7 +196,7 @@ public:
 			truncated = false;
 			if(args.contains("truncated"))
 			{
-				if(args["truncated"].type() != QVariant::Bool)
+				if(typeId(args["truncated"]) != QMetaType::Bool)
 				{
 					respondError("bad-request");
 					return;
@@ -206,7 +208,7 @@ public:
 			bool getSession = false;
 			if(args.contains("get-session"))
 			{
-				if(args["get-session"].type() != QVariant::Bool)
+				if(typeId(args["get-session"]) != QMetaType::Bool)
 				{
 					respondError("bad-request");
 					return;
@@ -218,7 +220,7 @@ public:
 			autoShare = false;
 			if(args.contains("auto-share"))
 			{
-				if(args["auto-share"].type() != QVariant::Bool)
+				if(typeId(args["auto-share"]) != QMetaType::Bool)
 				{
 					respondError("bad-request");
 					return;
@@ -471,7 +473,7 @@ public:
 		// process conn-max packets before doing anything else
 		if(args.contains("conn-max"))
 		{
-			if(args["conn-max"].type() != QVariant::List)
+			if(typeId(args["conn-max"]) != QMetaType::QVariantList)
 			{
 				respondError("bad-request");
 				return;
@@ -494,7 +496,7 @@ public:
 
 		if(args.contains("route"))
 		{
-			if(args["route"].type() != QVariant::ByteArray)
+			if(typeId(args["route"]) != QMetaType::QByteArray)
 			{
 				respondError("bad-request");
 				return;
@@ -505,7 +507,7 @@ public:
 
 		if(args.contains("separate-stats"))
 		{
-			if(args["separate-stats"].type() != QVariant::Bool)
+			if(typeId(args["separate-stats"]) != QMetaType::Bool)
 			{
 				respondError("bad-request");
 				return;
@@ -519,7 +521,7 @@ public:
 
 		if(args.contains("channel-prefix"))
 		{
-			if(args["channel-prefix"].type() != QVariant::ByteArray)
+			if(typeId(args["channel-prefix"]) != QMetaType::QByteArray)
 			{
 				respondError("bad-request");
 				return;
@@ -530,7 +532,7 @@ public:
 
 		if(args.contains("channels"))
 		{
-			if(args["channels"].type() != QVariant::List)
+			if(typeId(args["channels"]) != QMetaType::QVariantList)
 			{
 				respondError("bad-request");
 				return;
@@ -539,7 +541,7 @@ public:
 			QVariantList vchannels = args["channels"].toList();
 			foreach(const QVariant &v, vchannels)
 			{
-				if(v.type() != QVariant::ByteArray)
+				if(typeId(v) != QMetaType::QByteArray)
 				{
 					respondError("bad-request");
 					return;
@@ -551,7 +553,7 @@ public:
 
 		if(args.contains("trusted"))
 		{
-			if(args["trusted"].type() != QVariant::Bool)
+			if(typeId(args["trusted"]) != QMetaType::Bool)
 			{
 				respondError("bad-request");
 				return;
@@ -562,7 +564,7 @@ public:
 
 		// parse requests
 
-		if(!args.contains("requests") || args["requests"].type() != QVariant::List)
+		if(!args.contains("requests") || typeId(args["requests"]) != QMetaType::QVariantList)
 		{
 			respondError("bad-request");
 			return;
@@ -600,7 +602,7 @@ public:
 
 		// parse response
 
-		if(!args.contains("response") || args["response"].type() != QVariant::Hash)
+		if(!args.contains("response") || typeId(args["response"]) != QMetaType::QVariantHash)
 		{
 			respondError("bad-request");
 			return;
@@ -608,7 +610,7 @@ public:
 
 		QVariantHash rd = args["response"].toHash();
 
-		if(!rd.contains("code") || !rd["code"].canConvert(QVariant::Int))
+		if(!rd.contains("code") || !canConvert(rd["code"], QMetaType::Int))
 		{
 			respondError("bad-request");
 			return;
@@ -616,7 +618,7 @@ public:
 
 		responseData.code = rd["code"].toInt();
 
-		if(!rd.contains("reason") || rd["reason"].type() != QVariant::ByteArray)
+		if(!rd.contains("reason") || typeId(rd["reason"]) != QMetaType::QByteArray)
 		{
 			respondError("bad-request");
 			return;
@@ -624,7 +626,7 @@ public:
 
 		responseData.reason = rd["reason"].toByteArray();
 
-		if(!rd.contains("headers") || rd["headers"].type() != QVariant::List)
+		if(!rd.contains("headers") || typeId(rd["headers"]) != QMetaType::QVariantList)
 		{
 			respondError("bad-request");
 			return;
@@ -632,14 +634,14 @@ public:
 
 		foreach(const QVariant &vheader, rd["headers"].toList())
 		{
-			if(vheader.type() != QVariant::List)
+			if(typeId(vheader) != QMetaType::QVariantList)
 			{
 				respondError("bad-request");
 				return;
 			}
 
 			QVariantList vlist = vheader.toList();
-			if(vlist.count() != 2 || vlist[0].type() != QVariant::ByteArray || vlist[1].type() != QVariant::ByteArray)
+			if(vlist.count() != 2 || typeId(vlist[0]) != QMetaType::QByteArray || typeId(vlist[1]) != QMetaType::QByteArray)
 			{
 				respondError("bad-request");
 				return;
@@ -648,7 +650,7 @@ public:
 			responseData.headers += HttpHeader(vlist[0].toByteArray(), vlist[1].toByteArray());
 		}
 
-		if(!rd.contains("body") || rd["body"].type() != QVariant::ByteArray)
+		if(!rd.contains("body") || typeId(rd["body"]) != QMetaType::QByteArray)
 		{
 			respondError("bad-request");
 			return;
@@ -658,7 +660,7 @@ public:
 
 		if(args.contains("inspect"))
 		{
-			if(args["inspect"].type() != QVariant::Hash)
+			if(typeId(args["inspect"]) != QMetaType::QVariantHash)
 			{
 				respondError("bad-request");
 				return;
@@ -666,7 +668,7 @@ public:
 
 			QVariantHash vinspect = args["inspect"].toHash();
 
-			if(!vinspect.contains("no-proxy") || vinspect["no-proxy"].type() != QVariant::Bool)
+			if(!vinspect.contains("no-proxy") || typeId(vinspect["no-proxy"]) != QMetaType::Bool)
 			{
 				respondError("bad-request");
 				return;
@@ -677,7 +679,7 @@ public:
 			inspectInfo.sharingKey.clear();
 			if(vinspect.contains("sharing-key"))
 			{
-				if(vinspect["sharing-key"].type() != QVariant::ByteArray)
+				if(typeId(vinspect["sharing-key"]) != QMetaType::QByteArray)
 				{
 					respondError("bad-request");
 					return;
@@ -688,7 +690,7 @@ public:
 
 			if(vinspect.contains("sid"))
 			{
-				if(vinspect["sid"].type() != QVariant::ByteArray)
+				if(typeId(vinspect["sid"]) != QMetaType::QByteArray)
 				{
 					respondError("bad-request");
 					return;
@@ -699,7 +701,7 @@ public:
 
 			if(vinspect.contains("last-ids"))
 			{
-				if(vinspect["last-ids"].type() != QVariant::Hash)
+				if(typeId(vinspect["last-ids"]) != QMetaType::QVariantHash)
 				{
 					respondError("bad-request");
 					return;
@@ -711,7 +713,7 @@ public:
 				{
 					it.next();
 
-					if(it.value().type() != QVariant::ByteArray)
+					if(typeId(it.value()) != QMetaType::QByteArray)
 					{
 						respondError("bad-request");
 						return;
@@ -730,7 +732,7 @@ public:
 
 		if(args.contains("response-sent"))
 		{
-			if(args["response-sent"].type() != QVariant::Bool)
+			if(typeId(args["response-sent"]) != QMetaType::Bool)
 			{
 				respondError("bad-request");
 				return;
@@ -742,7 +744,7 @@ public:
 		bool useSession = false;
 		if(args.contains("use-session"))
 		{
-			if(args["use-session"].type() != QVariant::Bool)
+			if(typeId(args["use-session"]) != QMetaType::Bool)
 			{
 				respondError("bad-request");
 				return;
@@ -818,47 +820,46 @@ public:
 		return out;
 	}
 
-signals:
-	void sessionsReady();
-	void retryPacketReady(const RetryRequestPacket &packet);
+	Signal sessionsReady;
+	boost::signals2::signal<void(const QByteArray &,const RetryRequestPacket&)> retryPacketReady;
 
 private:
 	static HttpRequestData parseRequestData(const QVariantHash &args, const QString &field)
 	{
-		if(!args.contains(field) || args[field].type() != QVariant::Hash)
+		if(!args.contains(field) || typeId(args[field]) != QMetaType::QVariantHash)
 			return HttpRequestData();
 
 		QVariantHash rd = args[field].toHash();
 
-		if(!rd.contains("method") || rd["method"].type() != QVariant::ByteArray)
+		if(!rd.contains("method") || typeId(rd["method"]) != QMetaType::QByteArray)
 			return HttpRequestData();
 
 		HttpRequestData out;
 		out.method = QString::fromLatin1(rd["method"].toByteArray());
 
-		if(!rd.contains("uri") || rd["uri"].type() != QVariant::ByteArray)
+		if(!rd.contains("uri") || typeId(rd["uri"]) != QMetaType::QByteArray)
 			return HttpRequestData();
 
 		out.uri = QUrl(rd["uri"].toString(), QUrl::StrictMode);
 		if(!out.uri.isValid())
 			return HttpRequestData();
 
-		if(!rd.contains("headers") || rd["headers"].type() != QVariant::List)
+		if(!rd.contains("headers") || typeId(rd["headers"]) != QMetaType::QVariantList)
 			return HttpRequestData();
 
 		foreach(const QVariant &vheader, rd["headers"].toList())
 		{
-			if(vheader.type() != QVariant::List)
+			if(typeId(vheader) != QMetaType::QVariantList)
 				return HttpRequestData();
 
 			QVariantList vlist = vheader.toList();
-			if(vlist.count() != 2 || vlist[0].type() != QVariant::ByteArray || vlist[1].type() != QVariant::ByteArray)
+			if(vlist.count() != 2 || typeId(vlist[0]) != QMetaType::QByteArray || typeId(vlist[1]) != QMetaType::QByteArray)
 				return HttpRequestData();
 
 			out.headers += HttpHeader(vlist[0].toByteArray(), vlist[1].toByteArray());
 		}
 
-		if(!rd.contains("body") || rd["body"].type() != QVariant::ByteArray)
+		if(!rd.contains("body") || typeId(rd["body"]) != QMetaType::QByteArray)
 			return HttpRequestData();
 
 		out.body = rd["body"].toByteArray();
@@ -958,11 +959,13 @@ private:
 			return;
 		}
 
+		QByteArray reqFrom = req->from();
+
 		QVariantHash result;
 		result["accepted"] = true;
 		req->respond(result);
 
-		log_debug("accepting %d requests", requestStates.count());
+		log_debug("accepting %d requests from %s", requestStates.count(), reqFrom.data());
 
 		if(instruct.holdMode == Instruct::ResponseHold)
 		{
@@ -996,7 +999,7 @@ private:
 
 					needRemoveFromStats.remove(cid);
 
-					int unreportedTime = stats->removeConnection(cid, true);
+					int unreportedTime = stats->removeConnection(cid, true, reqFrom);
 
 					RetryRequestPacket::Request rpreq;
 					rpreq.rid = rs.rid;
@@ -1045,9 +1048,9 @@ private:
 				}
 
 				rp.route = route.toUtf8();
-				rp.retrySeq = stats->lastRetrySeq();
+				rp.retrySeq = stats->lastRetrySeq(reqFrom);
 
-				emit retryPacketReady(rp);
+				retryPacketReady(reqFrom, rp);
 
 				setFinished(true);
 				return;
@@ -1086,6 +1089,7 @@ private:
 				implicitChannelsSet += channel;
 
 			HttpSession::AcceptData adata;
+			adata.from = reqFrom;
 			adata.requestData = origRequestData;
 			adata.logicalPeerAddress = rs.logicalPeerAddress;
 			adata.debug = rs.debug;
@@ -1112,7 +1116,7 @@ private:
 
 		// engine should directly connect to this and register the holds
 		//   immediately, to avoid a race with the lastId check
-		emit sessionsReady();
+		sessionsReady();
 
 		setFinished(true);
 	}
@@ -1170,8 +1174,7 @@ public:
 		timer_->start(SUBSCRIBED_DELAY);
 	}
 
-signals:
-	void subscribed();
+	Signal subscribed;
 
 private:
 	QString channel_;
@@ -1180,7 +1183,7 @@ private:
 private slots:
 	void timer_timeout()
 	{
-		emit subscribed();
+		subscribed();
 	}
 };
 
@@ -1215,6 +1218,12 @@ public:
 		}
 	};
 
+	struct WSSessionConnections {
+		Connection sendConnection;
+		Connection expConnection;
+		Connection errorConnection;
+	};
+
 	HandlerEngine *q;
 	Configuration config;
 	ZhttpManager *zhttpIn;
@@ -1229,16 +1238,17 @@ public:
 	QZmq::Socket *inSubSock;
 	QZmq::Valve *inSubValve;
 	QZmq::Socket *retrySock;
-	QZmq::Socket *wsControlInSock;
-	QZmq::Valve *wsControlInValve;
-	QZmq::Socket *wsControlOutSock;
+	QZmq::Socket *wsControlInitSock;
+	QZmq::Valve *wsControlInitValve;
+	QZmq::Socket *wsControlStreamSock;
+	QZmq::Valve *wsControlStreamValve;
 	QZmq::Socket *statsSock;
 	QZmq::Socket *proxyStatsSock;
 	QZmq::Valve *proxyStatsValve;
 	SimpleHttpServer *controlHttpServer;
 	StatsManager *stats;
-	RateLimiter *publishLimiter;
-	RateLimiter *updateLimiter;
+	std::unique_ptr<RateLimiter> publishLimiter;
+	std::unique_ptr<RateLimiter> updateLimiter;
 	HttpSessionUpdateManager *httpSessionUpdateManager;
 	Sequencer *sequencer;
 	CommonState cs;
@@ -1252,6 +1262,18 @@ public:
 	Connection controlServerConnection;
 	Connection itemReadyConnection;
 	map<Deferred*, Connection> finishedConnection;
+	map<Subscription*, Connection> subscribedConnection;
+	map<AcceptWorker*, Connection> retryPacketReadyConnection;
+	map<AcceptWorker*, Connection> sessionsReadyConnection;
+	Connection connectionsRefreshedConnection;
+	Connection unsubscribedConnection;
+	Connection reportedConnection;
+	map<WsSession*, WSSessionConnections> wsSessionConnectionMap;
+	Connection pullConnection;
+	Connection controlInitValveConnection;
+	Connection controlStreamValveConnection;
+	Connection inSubValveConnection;
+	Connection proxyStatConnection;
 
 	Private(HandlerEngine *_q) :
 		QObject(_q),
@@ -1268,9 +1290,10 @@ public:
 		inSubSock(0),
 		inSubValve(0),
 		retrySock(0),
-		wsControlInSock(0),
-		wsControlInValve(0),
-		wsControlOutSock(0),
+		wsControlInitSock(0),
+		wsControlInitValve(0),
+		wsControlStreamSock(0),
+		wsControlStreamValve(0),
 		statsSock(0),
 		proxyStatsSock(0),
 		proxyStatsValve(0),
@@ -1280,8 +1303,8 @@ public:
 	{
 		qRegisterMetaType<DetectRuleList>();
 
-		publishLimiter = new RateLimiter(this);
-		updateLimiter = new RateLimiter(this);
+		publishLimiter = std::make_unique<RateLimiter>();
+		updateLimiter = std::make_unique<RateLimiter>();
 
 		httpSessionUpdateManager = new HttpSessionUpdateManager(this);
 
@@ -1329,36 +1352,36 @@ public:
 		log_info("zhttp in stream: %s", qPrintable(config.serverInStreamSpecs.join(", ")));
 		log_info("zhttp out: %s", qPrintable(config.serverOutSpecs.join(", ")));
 
-		if(!config.inspectSpec.isEmpty())
+		if(!config.inspectSpecs.isEmpty())
 		{
 			inspectServer = new ZrpcManager(this);
 			inspectServer->setBind(false);
 			inspectServer->setIpcFileMode(config.ipcFileMode);
 			inspectReqReadyConnection = inspectServer->requestReady.connect(boost::bind(&Private::inspectServer_requestReady, this));
 
-			if(!inspectServer->setServerSpecs(QStringList() << config.inspectSpec))
+			if(!inspectServer->setServerSpecs(config.inspectSpecs))
 			{
 				// zrpcmanager logs error
 				return false;
 			}
 
-			log_info("inspect server: %s", qPrintable(config.inspectSpec));
+			log_info("inspect server: %s", qPrintable(config.inspectSpecs.join(", ")));
 		}
 
-		if(!config.acceptSpec.isEmpty())
+		if(!config.acceptSpecs.isEmpty())
 		{
 			acceptServer = new ZrpcManager(this);
 			acceptServer->setBind(false);
 			acceptServer->setIpcFileMode(config.ipcFileMode);
 			acceptReqReadyConnection = acceptServer->requestReady.connect(boost::bind(&Private::acceptServer_requestReady, this));
 
-			if(!acceptServer->setServerSpecs(QStringList() << config.acceptSpec))
+			if(!acceptServer->setServerSpecs(config.acceptSpecs))
 			{
 				// zrpcmanager logs error
 				return false;
 			}
 
-			log_info("accept server: %s", qPrintable(config.acceptSpec));
+			log_info("accept server: %s", qPrintable(config.acceptSpecs.join(", ")));
 		}
 
 		if(!config.stateSpec.isEmpty())
@@ -1401,12 +1424,12 @@ public:
 			QString errorMessage;
 			if(!ZUtil::setupSocket(inPullSock, config.pushInSpec, true, config.ipcFileMode, &errorMessage))
 			{
-					log_error("%s", qPrintable(errorMessage));
-					return false;
+				log_error("%s", qPrintable(errorMessage));
+				return false;
 			}
 
 			inPullValve = new QZmq::Valve(inPullSock, this);
-			connect(inPullValve, &QZmq::Valve::readyRead, this, &Private::inPull_readyRead);
+			pullConnection = inPullValve->readyRead.connect(boost::bind(&Private::inPull_readyRead, this, boost::placeholders::_1));
 
 			log_info("in pull: %s", qPrintable(config.pushInSpec));
 		}
@@ -1420,8 +1443,8 @@ public:
 			QString errorMessage;
 			if(!ZUtil::setupSocket(inSubSock, config.pushInSubSpecs, !config.pushInSubConnect, config.ipcFileMode, &errorMessage))
 			{
-					log_error("%s", qPrintable(errorMessage));
-					return false;
+				log_error("%s", qPrintable(errorMessage));
+				return false;
 			}
 
 			if(config.pushInSubConnect)
@@ -1433,61 +1456,78 @@ public:
 			}
 
 			inSubValve = new QZmq::Valve(inSubSock, this);
-			connect(inSubValve, &QZmq::Valve::readyRead, this, &Private::inSub_readyRead);
+			inSubValveConnection = inSubValve->readyRead.connect(boost::bind(&Private::inSub_readyRead, this, boost::placeholders::_1));
 
 			log_info("in sub: %s", qPrintable(config.pushInSubSpecs.join(", ")));
 		}
 
-		if(!config.retryOutSpec.isEmpty())
+		if(!config.retryOutSpecs.isEmpty())
 		{
-			retrySock = new QZmq::Socket(QZmq::Socket::Push, this);
+			retrySock = new QZmq::Socket(QZmq::Socket::Router, this);
+			retrySock->setImmediateEnabled(true);
 			retrySock->setHwm(DEFAULT_HWM);
 			retrySock->setShutdownWaitTime(RETRY_WAIT_TIME);
+			retrySock->setRouterMandatoryEnabled(true);
 
-			QString errorMessage;
-			if(!ZUtil::setupSocket(retrySock, config.retryOutSpec, false, config.ipcFileMode, &errorMessage))
+			foreach(const QString &spec, config.retryOutSpecs)
 			{
+				QString errorMessage;
+				if(!ZUtil::setupSocket(retrySock, spec, false, config.ipcFileMode, &errorMessage))
+				{
 					log_error("%s", qPrintable(errorMessage));
 					return false;
+				}
 			}
 
-			log_info("retry: %s", qPrintable(config.retryOutSpec));
+			log_info("retry: %s", qPrintable(config.retryOutSpecs.join(", ")));
 		}
 
-		if(!config.wsControlInSpec.isEmpty() && !config.wsControlOutSpec.isEmpty())
+		if(!config.wsControlInitSpecs.isEmpty() && !config.wsControlStreamSpecs.isEmpty())
 		{
-			wsControlInSock = new QZmq::Socket(QZmq::Socket::Pull, this);
-			wsControlInSock->setHwm(DEFAULT_HWM);
+			wsControlInitSock = new QZmq::Socket(QZmq::Socket::Pull, this);
+			wsControlInitSock->setHwm(DEFAULT_HWM);
 
-			QString errorMessage;
-			if(!ZUtil::setupSocket(wsControlInSock, config.wsControlInSpec, false, config.ipcFileMode, &errorMessage))
+			foreach(const QString &spec, config.wsControlInitSpecs)
 			{
+				QString errorMessage;
+				if(!ZUtil::setupSocket(wsControlInitSock, spec, false, config.ipcFileMode, &errorMessage))
+				{
 					log_error("%s", qPrintable(errorMessage));
 					return false;
+				}
 			}
 
-			wsControlInValve = new QZmq::Valve(wsControlInSock, this);
-			connect(wsControlInValve, &QZmq::Valve::readyRead, this, &Private::wsControlIn_readyRead);
+			wsControlInitValve = new QZmq::Valve(wsControlInitSock, this);
+			controlInitValveConnection = wsControlInitValve->readyRead.connect(boost::bind(&Private::wsControlInit_readyRead, this, boost::placeholders::_1));
 
-			log_info("ws control in: %s", qPrintable(config.wsControlInSpec));
+			log_info("ws control init: %s", qPrintable(config.wsControlInitSpecs.join(", ")));
 
-			wsControlOutSock = new QZmq::Socket(QZmq::Socket::Push, this);
-			wsControlOutSock->setHwm(DEFAULT_HWM);
-			wsControlOutSock->setShutdownWaitTime(WSCONTROL_WAIT_TIME);
+			wsControlStreamSock = new QZmq::Socket(QZmq::Socket::Router, this);
+			wsControlStreamSock->setIdentity(config.instanceId);
+			wsControlStreamSock->setImmediateEnabled(true);
+			wsControlStreamSock->setHwm(DEFAULT_HWM);
+			wsControlStreamSock->setShutdownWaitTime(WSCONTROL_WAIT_TIME);
 
-			if(!ZUtil::setupSocket(wsControlOutSock, config.wsControlOutSpec, false, config.ipcFileMode, &errorMessage))
+			foreach(const QString &spec, config.wsControlStreamSpecs)
 			{
+				QString errorMessage;
+				if(!ZUtil::setupSocket(wsControlStreamSock, spec, false, config.ipcFileMode, &errorMessage))
+				{
 					log_error("%s", qPrintable(errorMessage));
 					return false;
+				}
 			}
 
-			log_info("ws control out: %s", qPrintable(config.wsControlOutSpec));
+			wsControlStreamValve = new QZmq::Valve(wsControlStreamSock, this);
+			controlStreamValveConnection = wsControlStreamValve->readyRead.connect(boost::bind(&Private::wsControlStream_readyRead, this, boost::placeholders::_1));
+
+			log_info("ws control stream: %s", qPrintable(config.wsControlStreamSpecs.join(", ")));
 		}
 
 		stats = new StatsManager(config.connectionsMax, config.connectionsMax * config.connectionSubscriptionMax, this);
-		connect(stats, &StatsManager::connectionsRefreshed, this, &Private::stats_connectionsRefreshed);
-		connect(stats, &StatsManager::unsubscribed, this, &Private::stats_unsubscribed);
-		connect(stats, &StatsManager::reported, this, &Private::stats_reported);
+		connectionsRefreshedConnection = stats->connectionsRefreshed.connect(boost::bind(&Private::stats_connectionsRefreshed, this, boost::placeholders::_1));
+		unsubscribedConnection = stats->unsubscribed.connect(boost::bind(&Private::stats_unsubscribed, this, boost::placeholders::_1, boost::placeholders::_2));
+		reportedConnection = stats->reported.connect(boost::bind(&Private::stats_reported, this, boost::placeholders::_1));
 
 		stats->setConnectionSendEnabled(config.statsConnectionSend);
 		stats->setConnectionTtl(config.statsConnectionTtl);
@@ -1529,24 +1569,27 @@ public:
 			}
 		}
 
-		if(!config.proxyStatsSpec.isEmpty())
+		if(!config.proxyStatsSpecs.isEmpty())
 		{
 			proxyStatsSock = new QZmq::Socket(QZmq::Socket::Sub, this);
 			proxyStatsSock->setHwm(DEFAULT_HWM);
 			proxyStatsSock->setShutdownWaitTime(0);
 			proxyStatsSock->subscribe("");
 
-			QString errorMessage;
-			if(!ZUtil::setupSocket(proxyStatsSock, config.proxyStatsSpec, false, config.ipcFileMode, &errorMessage))
+			foreach(const QString &spec, config.proxyStatsSpecs)
 			{
-					log_error("%s", qPrintable(errorMessage));
-					return false;
+				QString errorMessage;
+				if(!ZUtil::setupSocket(proxyStatsSock, spec, false, config.ipcFileMode, &errorMessage))
+				{
+						log_error("%s", qPrintable(errorMessage));
+						return false;
+				}
 			}
 
 			proxyStatsValve = new QZmq::Valve(proxyStatsSock, this);
-			connect(proxyStatsValve, &QZmq::Valve::readyRead, this, &Private::proxyStats_readyRead);
+			proxyStatConnection = proxyStatsValve->readyRead.connect(boost::bind(&Private::proxyStats_readyRead, this, boost::placeholders::_1));
 
-			log_info("proxy stats: %s", qPrintable(config.proxyStatsSpec));
+			log_info("proxy stats: %s", qPrintable(config.proxyStatsSpecs.join(", ")));
 		}
 
 		if(!config.proxyCommandSpec.isEmpty())
@@ -1577,8 +1620,10 @@ public:
 			inPullValve->open();
 		if(inSubValve)
 			inSubValve->open();
-		if(wsControlInValve)
-			wsControlInValve->open();
+		if(wsControlInitValve)
+			wsControlInitValve->open();
+		if(wsControlStreamValve)
+			wsControlStreamValve->open();
 		if(proxyStatsValve)
 			proxyStatsValve->open();
 
@@ -1602,7 +1647,7 @@ private:
 		sequencer->addItem(item, seq);
 	}
 
-	void writeRetryPacket(const RetryRequestPacket &packet)
+	void writeRetryPacket(const QByteArray &instanceAddress, const RetryRequestPacket &packet)
 	{
 		if(!retrySock)
 		{
@@ -1613,28 +1658,37 @@ private:
 		QVariant vout = packet.toVariant();
 
 		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
-			log_debug("OUT retry: %s", qPrintable(TnetString::variantToString(vout, -1)));
+			log_debug("OUT retry: to=%s %s", instanceAddress.data(), qPrintable(TnetString::variantToString(vout, -1)));
 
-		retrySock->write(QList<QByteArray>() << TnetString::fromVariant(vout));
+		QList<QByteArray> msg;
+		msg += instanceAddress;
+		msg += QByteArray();
+		msg += TnetString::fromVariant(vout);
+		retrySock->write(msg);
 	}
 
-	void writeWsControlItems(const QList<WsControlPacket::Item> &items)
+	void writeWsControlItems(const QByteArray &instanceAddress, const QList<WsControlPacket::Item> &items)
 	{
-		if(!wsControlOutSock)
+		if(!wsControlStreamSock)
 		{
 			log_error("wscontrol: can't write, no socket");
 			return;
 		}
 
 		WsControlPacket out;
+		out.from = config.instanceId;
 		out.items = items;
 
 		QVariant vout = out.toVariant();
 
 		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
-			log_debug("OUT wscontrol: %s", qPrintable(TnetString::variantToString(vout, -1)));
+			log_debug("OUT wscontrol: to=%s %s", instanceAddress.data(), qPrintable(TnetString::variantToString(vout, -1)));
 
-		wsControlOutSock->write(QList<QByteArray>() << TnetString::fromVariant(vout));
+		QList<QByteArray> msg;
+		msg += instanceAddress;
+		msg += QByteArray();
+		msg += TnetString::fromVariant(vout);
+		wsControlStreamSock->write(msg);
 	}
 
 	void addSub(const QString &channel)
@@ -1642,7 +1696,7 @@ private:
 		if(!cs.subs.contains(channel))
 		{
 			Subscription *sub = new Subscription(channel);
-			connect(sub, &Subscription::subscribed, this, &Private::sub_subscribed);
+			subscribedConnection[sub] = sub->subscribed.connect(boost::bind(&Private::sub_subscribed, this, sub));
 			cs.subs.insert(channel, sub);
 			sub->start();
 
@@ -1660,6 +1714,7 @@ private:
 		{
 			Subscription *sub = cs.subs[channel];
 			cs.subs.remove(channel);
+			subscribedConnection.erase(sub);
 			delete sub;
 
 			sequencer->clearPendingForChannel(channel);
@@ -1680,6 +1735,7 @@ private:
 		log_debug("removed ws session: %s", qPrintable(s->cid));
 
 		cs.wsSessions.remove(s->cid);
+		wsSessionConnectionMap.erase(s);
 		delete s;
 	}
 
@@ -1778,13 +1834,10 @@ private:
 			}
 			else if(f.action == PublishFormat::Refresh)
 			{
-				Deferred *d = ControlRequest::refresh(proxyControlClient, i.cid, this);
-				finishedConnection[d] = d->finished.connect(boost::bind(&Private::deferred_finished, this, boost::placeholders::_1, d));
-				deferreds += d;
-				return;
+				i.type = WsControlPacket::Item::Refresh;
 			}
 
-			writeWsControlItems(QList<WsControlPacket::Item>() << i);
+			writeWsControlItems(s->peer, QList<WsControlPacket::Item>() << i);
 		}
 	}
 
@@ -1944,10 +1997,10 @@ private:
 			// accept request immediately before returning to the event loop.
 			// the start() call will do this
 
-			AcceptWorker *w = new AcceptWorker(req, stateClient, &cs, zhttpIn, zhttpOut, stats, updateLimiter, httpSessionUpdateManager, config.connectionSubscriptionMax, this);
+			AcceptWorker *w = new AcceptWorker(req, stateClient, &cs, zhttpIn, zhttpOut, stats, updateLimiter.get(), httpSessionUpdateManager, config.connectionSubscriptionMax, this);
 			finishedConnection[w] = w->finished.connect(boost::bind(&Private::acceptWorker_finished, this, boost::placeholders::_1, w));
-			connect(w, &AcceptWorker::sessionsReady, this, &Private::acceptWorker_sessionsReady);
-			connect(w, &AcceptWorker::retryPacketReady, this, &Private::acceptWorker_retryPacketReady);
+			sessionsReadyConnection[w] = w->sessionsReady.connect(boost::bind(&Private::acceptWorker_sessionsReady, this, w));
+			retryPacketReadyConnection[w] =  w->retryPacketReady.connect(boost::bind(&Private::acceptWorker_retryPacketReady, this, boost::placeholders::_1, boost::placeholders::_2));
 			acceptWorkers += w;
 
 			w->start();
@@ -1958,7 +2011,7 @@ private:
 
 			if(args.contains("conn-max"))
 			{
-				if(args["conn-max"].type() == QVariant::List)
+				if(typeId(args["conn-max"]) == QMetaType::QVariantList)
 				{
 					QVariantList packets = args["conn-max"].toList();
 
@@ -2032,7 +2085,7 @@ private:
 				return;
 			}
 
-			if(args["items"].type() != QVariant::List)
+			if(typeId(args["items"]) != QMetaType::QVariantList)
 			{
 				req->respondError("bad-request", "Invalid format: object contains 'items' with wrong type");
 				delete req;
@@ -2319,6 +2372,8 @@ private:
 		Q_UNUSED(result);
 
 		finishedConnection.erase(w);
+		sessionsReadyConnection.erase(w);
+		retryPacketReadyConnection.erase(w);
 		acceptWorkers.remove(w);
 
 		// try to read again
@@ -2333,7 +2388,99 @@ private:
 		deferreds.remove(w);
 	}
 	
-private slots:
+	void sub_subscribed(Subscription *sub)
+	{
+		updateSessions(sub->channel());
+	}
+
+	void acceptWorker_sessionsReady(AcceptWorker *w)
+	{
+		QList<HttpSession*> sessions = w->takeSessions();
+		foreach(HttpSession *hs, sessions)
+		{
+			// NOTE: for performance reasons we do not call hs->setParent and
+			// instead leave the object unparented
+
+			hs->subscribeCallback().add(Private::hs_subscribe_cb, this);
+			hs->unsubscribeCallback().add(Private::hs_unsubscribe_cb, this);
+			hs->finishedCallback().add(Private::hs_finished_cb, this);
+
+			cs.httpSessions.insert(hs->rid(), hs);
+
+			hs->start();
+		}
+	}
+
+	void acceptWorker_retryPacketReady(const QByteArray &instanceAddress, const RetryRequestPacket &packet)
+	{
+		writeRetryPacket(instanceAddress, packet);
+	}
+
+	void stats_connectionsRefreshed(const QList<QByteArray> &ids)
+	{
+		if(stateClient)
+		{
+			// find sids of the connections
+			QHash<QString, LastIds> sidLastIds;
+			foreach(const QByteArray &id, ids)
+			{
+				int at = id.indexOf(':');
+				assert(at != -1);
+				ZhttpRequest::Rid rid(id.mid(0, at), id.mid(at + 1));
+
+				HttpSession *hs = cs.httpSessions.value(rid);
+				if(hs && !hs->sid().isEmpty())
+					sidLastIds[hs->sid()] = LastIds();
+			}
+
+			if(!sidLastIds.isEmpty())
+			{
+				Deferred *d = SessionRequest::updateMany(stateClient, sidLastIds, this);
+				finishedConnection[d] = d->finished.connect(boost::bind(&Private::sessionUpdateMany_finished, this, boost::placeholders::_1, d));
+				deferreds += d;
+			}
+		}
+	}
+
+	void stats_unsubscribed(const QString &mode, const QString &channel)
+	{
+		// NOTE: this callback may be invoked while looping over certain structures,
+		//   so be careful what you touch
+
+		Q_UNUSED(mode);
+
+		if(!cs.responseSessionsByChannel.contains(channel) && !cs.streamSessionsByChannel.contains(channel) && !cs.wsSessionsByChannel.contains(channel))
+			removeSub(channel);
+	}
+
+	void stats_reported(const QList<StatsPacket> &packets)
+	{
+		// only one outstanding report at a time
+		if(report)
+			return;
+
+		// consolidate data
+		StatsPacket all;
+		all.type = StatsPacket::Report;
+		all.connectionsMax = 0;
+		all.connectionsMinutes = 0;
+		all.messagesReceived = 0;
+		all.messagesSent = 0;
+		all.httpResponseMessagesSent = 0;
+		foreach(const StatsPacket &p, packets)
+		{
+			all.connectionsMax += qMax(p.connectionsMax, 0);
+			all.connectionsMinutes += qMax(p.connectionsMinutes, 0);
+			all.messagesReceived += qMax(p.messagesReceived, 0);
+			all.messagesSent += qMax(p.messagesSent, 0);
+			all.httpResponseMessagesSent += qMax(p.httpResponseMessagesSent, 0);
+		}
+
+		report = ControlRequest::report(proxyControlClient, all, this);
+		finishedConnection[report] = report->finished.connect(boost::bind(&Private::report_finished, this, boost::placeholders::_1));
+		deferreds += report;
+	}
+
 	QVariant parseJsonOrTnetstring(const QByteArray &message, bool *ok = 0, QString *errorMessage = 0) {
 		QVariant data;
 		bool ok_;
@@ -2445,7 +2592,7 @@ private slots:
 		handlePublishItem(item);
 	}
 
-	void wsControlIn_readyRead(const QList<QByteArray> &message)
+	void wsControlInit_readyRead(const QList<QByteArray> &message)
 	{
 		if(message.count() != 1)
 		{
@@ -2453,8 +2600,26 @@ private slots:
 			return;
 		}
 
+		wsControlIn_readyRead(message[0]);
+	}
+
+	void wsControlStream_readyRead(const QList<QByteArray> &message)
+	{
+		QZmq::ReqMessage req(message);
+
+		if(req.content().count() != 1)
+		{
+			log_warning("IN wscontrol: received message with parts != 1, skipping");
+			return;
+		}
+
+		wsControlIn_readyRead(req.content()[0]);
+	}
+
+	void wsControlIn_readyRead(const QByteArray &message)
+	{
 		bool ok;
-		QVariant data = TnetString::toVariant(message[0], 0, &ok);
+		QVariant data = TnetString::toVariant(message, 0, &ok);
 		if(!ok)
 		{
 			log_warning("IN wscontrol: received message with invalid format (tnetstring parse failed), skipping");
@@ -2494,9 +2659,12 @@ private slots:
 				if(!s)
 				{
 					s = new WsSession(this);
-					connect(s, &WsSession::send, this, &Private::wssession_send);
-					connect(s, &WsSession::expired, this, &Private::wssession_expired);
-					connect(s, &WsSession::error, this, &Private::wssession_error);
+					wsSessionConnectionMap[s] = {
+						s->send.connect(boost::bind(&Private::wssession_send, this, boost::placeholders::_1, boost::placeholders::_2, boost::placeholders::_3, s)),
+						s->expired.connect(boost::bind(&Private::wssession_expired, this, s)),
+						s->error.connect(boost::bind(&Private::wssession_error, this, s))
+					};
+					s->peer = packet.from;
 					s->cid = QString::fromUtf8(item.cid);
 					s->ttl = item.ttl;
 					s->requestData.uri = item.uri;
@@ -2721,7 +2889,7 @@ private slots:
 		}
 
 		if(!outItems.isEmpty())
-			writeWsControlItems(outItems);
+			writeWsControlItems(packet.from, outItems);
 
 		if(stateClient)
 		{
@@ -2895,7 +3063,7 @@ private slots:
 					return;
 				}
 
-				if(mdata["items"].type() != QVariant::List)
+				if(typeId(mdata["items"]) != QMetaType::QVariantList)
 				{
 					httpControlRespond(req, 400, "Bad Request", "Invalid format: object contains 'items' with wrong type\n");
 					return;
@@ -2967,31 +3135,7 @@ private slots:
 		}
 	}
 
-	void acceptWorker_sessionsReady()
-	{
-		AcceptWorker *w = (AcceptWorker *)sender();
-
-		QList<HttpSession*> sessions = w->takeSessions();
-		foreach(HttpSession *hs, sessions)
-		{
-			// NOTE: for performance reasons we do not call hs->setParent and
-			// instead leave the object unparented
-
-			hs->subscribeCallback().add(Private::hs_subscribe_cb, this);
-			hs->unsubscribeCallback().add(Private::hs_unsubscribe_cb, this);
-			hs->finishedCallback().add(Private::hs_finished_cb, this);
-
-			cs.httpSessions.insert(hs->rid(), hs);
-
-			hs->start();
-		}
-	}
-
-	void acceptWorker_retryPacketReady(const RetryRequestPacket &packet)
-	{
-		writeRetryPacket(packet);
-	}
-
+private slots:
 	void hs_subscribe(HttpSession *hs, const QString &channel)
 	{
 		Instruct::HoldMode mode = hs->holdMode();
@@ -3037,6 +3181,7 @@ private slots:
 
 	void hs_finished(HttpSession *hs)
 	{
+		QByteArray addr = hs->retryToAddress();
 		RetryRequestPacket rp = hs->retryPacket();
 
 		cs.httpSessions.remove(hs->rid());
@@ -3047,13 +3192,11 @@ private slots:
 		hs->deleteLater();
 
 		if(!rp.requests.isEmpty())
-			writeRetryPacket(rp);
+			writeRetryPacket(addr, rp);
 	}
 
-	void wssession_send(int reqId, const QByteArray &type, const QByteArray &message)
+	void wssession_send(int reqId, const QByteArray &type, const QByteArray &message, WsSession *s)
 	{
-		WsSession *s = (WsSession *)sender();
-
 		WsControlPacket::Item i;
 		i.cid = s->cid.toUtf8();
 		i.requestId = QByteArray::number(reqId);
@@ -3062,102 +3205,26 @@ private slots:
 		i.message = message;
 		i.queue = true;
 
-		writeWsControlItems(QList<WsControlPacket::Item>() << i);
+		writeWsControlItems(s->peer, QList<WsControlPacket::Item>() << i);
 	}
 
-	void wssession_expired()
+	void wssession_expired(WsSession *s)
 	{
-		WsSession *s = (WsSession *)sender();
-
 		removeWsSession(s);
 	}
 
-	void wssession_error()
+	void wssession_error(WsSession *s)
 	{
-		WsSession *s = (WsSession *)sender();
-
 		log_debug("ws session %s control error", qPrintable(s->cid));
 
 		WsControlPacket::Item i;
 		i.cid = s->cid.toUtf8();
 		i.type = WsControlPacket::Item::Cancel;
 
-		writeWsControlItems(QList<WsControlPacket::Item>() << i);
+		writeWsControlItems(s->peer, QList<WsControlPacket::Item>() << i);
 
 		removeWsSession(s);
 	}
-
-	void sub_subscribed()
-	{
-		Subscription *sub = (Subscription *)sender();
-
-		updateSessions(sub->channel());
-	}
-
-	void stats_connectionsRefreshed(const QList<QByteArray> &ids)
-	{
-		if(stateClient)
-		{
-			// find sids of the connections
-			QHash<QString, LastIds> sidLastIds;
-			foreach(const QByteArray &id, ids)
-			{
-				int at = id.indexOf(':');
-				assert(at != -1);
-				ZhttpRequest::Rid rid(id.mid(0, at), id.mid(at + 1));
-
-				HttpSession *hs = cs.httpSessions.value(rid);
-				if(hs && !hs->sid().isEmpty())
-					sidLastIds[hs->sid()] = LastIds();
-			}
-
-			if(!sidLastIds.isEmpty())
-			{
-				Deferred *d = SessionRequest::updateMany(stateClient, sidLastIds, this);
-				finishedConnection[d] = d->finished.connect(boost::bind(&Private::sessionUpdateMany_finished, this, boost::placeholders::_1, d));
-				deferreds += d;
-			}
-		}
-	}
-
-	void stats_unsubscribed(const QString &mode, const QString &channel)
-	{
-		// NOTE: this callback may be invoked while looping over certain structures,
-		//   so be careful what you touch
-
-		Q_UNUSED(mode);
-
-		if(!cs.responseSessionsByChannel.contains(channel) && !cs.streamSessionsByChannel.contains(channel) && !cs.wsSessionsByChannel.contains(channel))
-			removeSub(channel);
-	}
-
-	void stats_reported(const QList<StatsPacket> &packets)
-	{
-		// only one outstanding report at a time
-		if(report)
-			return;
-
-		// consolidate data
-		StatsPacket all;
-		all.type = StatsPacket::Report;
-		all.connectionsMax = 0;
-		all.connectionsMinutes = 0;
-		all.messagesReceived = 0;
-		all.messagesSent = 0;
-		all.httpResponseMessagesSent = 0;
-		foreach(const StatsPacket &p, packets)
-		{
-			all.connectionsMax += qMax(p.connectionsMax, 0);
-			all.connectionsMinutes += qMax(p.connectionsMinutes, 0);
-			all.messagesReceived += qMax(p.messagesReceived, 0);
-			all.messagesSent += qMax(p.messagesSent, 0);
-			all.httpResponseMessagesSent += qMax(p.httpResponseMessagesSent, 0);
-		}
-
-		report = ControlRequest::report(proxyControlClient, all, this);
-		finishedConnection[report] = report->finished.connect(boost::bind(&Private::report_finished, this, boost::placeholders::_1));
-		deferreds += report;
-	}
 };
 
 HandlerEngine::HandlerEngine(QObject *parent) :
