@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2016-2023 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -29,6 +29,7 @@
 #include <QJsonObject>
 #include <QJsonArray>
 #include <QRandomGenerator>
+#include "qtcompat.h"
 #include "rtimer.h"
 #include "log.h"
 #include "bufferlist.h"
@@ -72,10 +73,10 @@ static QByteArray applyBodyPatch(const QByteArray &in, const QVariantList &bodyP
 		vbody = JsonPatch::patch(vbody, bodyPatch, &errorMessage);
 		if(vbody.isValid())
 			vbody = VariantUtil::convertToJsonStyle(vbody);
-		if(vbody.isValid() && (vbody.type() == QVariant::Map || vbody.type() == QVariant::List))
+		if(vbody.isValid() && (typeId(vbody) == QMetaType::QVariantMap || typeId(vbody) == QMetaType::QVariantList))
 		{
 			QJsonDocument doc;
-			if(vbody.type() == QVariant::Map)
+			if(typeId(vbody) == QMetaType::QVariantMap)
 				doc = QJsonDocument(QJsonObject::fromVariantMap(vbody.toMap()));
 			else // List
 				doc = QJsonDocument(QJsonArray::fromVariantList(vbody.toList()));
@@ -167,6 +168,7 @@ public:
 	Priority needUpdatePriority;
 	UpdateAction *pendingAction;
 	QList<PublishItem> publishQueue;
+	QByteArray retryToAddress;
 	RetryRequestPacket retryPacket;
 	LogUtil::Config logConfig;
 	FilterStack *responseFilters;
@@ -183,7 +185,7 @@ public:
 	Connection readyReadOutConnection;
 	Connection errorOutConnection;
 	Connection timerConnection;
-	Connection retryTimerConneciton;
+	Connection retryTimerConnection;
 
 	Private(HttpSession *_q, ZhttpRequest *_req, const HttpSession::AcceptData &_adata, const Instruct &_instruct, ZhttpManager *_outZhttp, StatsManager *_stats, RateLimiter *_updateLimiter, PublishLastIds *_publishLastIds, HttpSessionUpdateManager *_updateManager, int _connectionSubscriptionMax) :
 		QObject(_q),
@@ -211,11 +213,11 @@ public:
 		writeBytesChangedConnection = req->writeBytesChanged.connect(boost::bind(&Private::req_writeBytesChanged, this));
 		errorConnection = req->error.connect(boost::bind(&Private::req_error, this));
 
-		timer = new RTimer(this);
+		timer = new RTimer;
 		timerConnection = timer->timeout.connect(boost::bind(&Private::timer_timeout, this));
 
-		retryTimer = new RTimer(this);
-		retryTimerConneciton = retryTimer->timeout.connect(boost::bind(&Private::retryTimer_timeout, this));
+		retryTimer = new RTimer;
+		retryTimerConnection = retryTimer->timeout.connect(boost::bind(&Private::retryTimer_timeout, this));
 		retryTimer->setSingleShot(true);
 
 		adata = _adata;
@@ -245,7 +247,7 @@ public:
 		timer->setParent(0);
 		timer->deleteLater();
 
-		retryTimerConneciton.disconnect();
+		retryTimerConnection.disconnect();
 		retryTimer->setParent(0);
 		retryTimer->deleteLater();
 	}
@@ -1080,7 +1082,7 @@ private:
 
 			needRemoveFromStats = false;
 
-			int unreportedTime = stats->removeConnection(cid, true);
+			int unreportedTime = stats->removeConnection(cid, true, adata.from);
 
 			ZhttpRequest::ServerState ss = req->serverState();
 
@@ -1136,8 +1138,9 @@ private:
 			}
 
 			rp.route = adata.route.toUtf8();
-			rp.retrySeq = stats->lastRetrySeq();
+			rp.retrySeq = stats->lastRetrySeq(adata.from);
 
+			retryToAddress = adata.from;
 			retryPacket = rp;
 		}
 		else
@@ -1350,7 +1353,8 @@ private:
 	{
 		LogUtil::RequestData rd;
 
-		if(!adata.route.isEmpty())
+		// only log route id if explicitly set
+		if(!adata.statsRoute.isEmpty())
 			rd.routeId = adata.route;
 
 		rd.status = LogUtil::Response;
@@ -1369,7 +1373,8 @@ private:
 	{
 		LogUtil::RequestData rd;
 
-		if(!adata.route.isEmpty())
+		// only log route id if explicitly set
+		if(!adata.statsRoute.isEmpty())
 			rd.routeId = adata.route;
 
 		rd.status = LogUtil::Error;
@@ -1599,6 +1604,11 @@ QHash<QString, QString> HttpSession::meta() const
 	return d->instruct.meta;
 }
 
+QByteArray HttpSession::retryToAddress() const
+{
+	return d->retryToAddress;
+}
+
 RetryRequestPacket HttpSession::retryPacket() const
 {
 	return d->retryPacket;
