@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2012-2022 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -30,12 +30,16 @@
 #include <QFile>
 #include <QFileInfo>
 #include "processquit.h"
+#include "rtimer.h"
 #include "log.h"
 #include "settings.h"
 #include "xffrule.h"
+#include "domainmap.h"
 #include "engine.h"
 #include "config.h"
 
+using Connection = boost::signals2::scoped_connection;
+
 static void trimlist(QStringList *list)
 {
 	for(int n = 0; n < list->count(); ++n)
@@ -68,6 +72,22 @@ static XffRule parse_xffRule(const QStringList &in)
 	return out;
 }
 
+static QString suffixSpec(const QString &s, int i)
+{
+	if(s.startsWith("ipc:"))
+		return s + QString("-%1").arg(i);
+
+	return s;
+}
+
+static QStringList suffixSpecs(const QStringList &l, int i)
+{
+	if(l.count() == 1 && l[0].startsWith("ipc:"))
+		return QStringList() << (l[0] + QString("-%1").arg(i));
+
+	return l;
+}
+
 enum CommandLineParseResult
 {
 	CommandLineOk,
@@ -162,6 +182,158 @@ static CommandLineParseResult parseCommandLine(QCommandLineParser *parser, ArgsD
 	return CommandLineOk;
 }
 
+class EngineWorker : public QObject
+{
+	Q_OBJECT
+
+public:
+	EngineWorker(const Engine::Configuration &config, DomainMap *domainMap) :
+		QObject(),
+		config_(config),
+		engine_(new Engine(domainMap, this))
+	{
+	}
+
+	Signal started;
+	Signal stopped;
+	Signal error;
+
+public slots:
+	void start()
+	{
+		if(!engine_->start(config_))
+		{
+			delete engine_;
+			engine_ = 0;
+
+			error();
+			return;
+		}
+
+		started();
+	}
+
+	void stop()
+	{
+		delete engine_;
+		engine_ = 0;
+
+		stopped();
+	}
+
+	void routesChanged()
+	{
+		if(engine_)
+			engine_->routesChanged();
+	}
+
+private:
+	Engine::Configuration config_;
+	Engine *engine_;
+};
+
+class EngineThread : public QThread
+{
+	Q_OBJECT
+
+public:
+	QMutex m;
+	QWaitCondition w;
+	Engine::Configuration config;
+	DomainMap *domainMap;
+	EngineWorker *worker;
+
+	EngineThread(const Engine::Configuration &_config, DomainMap *_domainMap) :
+		config(_config),
+		domainMap(_domainMap),
+		worker(0)
+	{
+	}
+
+	~EngineThread()
+	{
+		stop();
+		wait();
+	}
+
+	bool start()
+	{
+		setObjectName("proxy-worker-" + QString::number(config.id));
+
+		QMutexLocker locker(&m);
+		QThread::start();
+		w.wait(&m);
+		return (bool)worker;
+	}
+
+	void stop()
+	{
+		QMutexLocker locker(&m);
+
+		if(worker)
+			QMetaObject::invokeMethod(worker, "stop", Qt::QueuedConnection);
+	}
+
+	void routesChanged()
+	{
+		QMutexLocker locker(&m);
+
+		if(worker)
+			QMetaObject::invokeMethod(worker, "routesChanged", Qt::QueuedConnection);
+	}
+
+	virtual void run()
+	{
+		// will unlock during exec
+		m.lock();
+
+		worker = new EngineWorker(config, domainMap);
+		Connection startedConnection = worker->started.connect(boost::bind(&EngineThread::worker_started, this));
+		Connection stoppedConnection = worker->stopped.connect(boost::bind(&EngineThread::worker_stopped, this));
+		Connection errorConnection = worker->error.connect(boost::bind(&EngineThread::worker_error, this));
+		QMetaObject::invokeMethod(worker, "start", Qt::QueuedConnection);
+		exec();
+
+		// ensure deferred deletes are processed
+		QCoreApplication::instance()->sendPostedEvents();
+
+		// deinit here, after all event loop activity has completed
+		RTimer::deinit();
+	}
+
+private:
+	void worker_started()
+	{
+		log_debug("worker %d: started", config.id);
+
+		// unblock start()
+		w.wakeOne();
+		m.unlock();
+	}
+
+	void worker_stopped()
+	{
+		delete worker;
+		worker = 0;
+
+		log_debug("worker %d: stopped", config.id);
+
+		quit();
+	}
+
+	void worker_error()
+	{
+		delete worker;
+		worker = 0;
+
+		quit();
+
+		// unblock start()
+		w.wakeOne();
+		m.unlock();
+	}
+};
+
 class App::Private : public QObject
 {
 	Q_OBJECT
@@ -169,14 +341,16 @@ class App::Private : public QObject
 public:
 	App *q;
 	ArgsData args;
-	Engine *engine;
+	DomainMap *domainMap;
+	std::list<EngineThread*> threads;
 	Connection quitConnection;
 	Connection hupConnection;
+	Connection changedConnection;
 
 	Private(App *_q) :
 		QObject(_q),
 		q(_q),
-		engine(0)
+		domainMap(0)
 	{
 		quitConnection = ProcessQuit::instance()->quit.connect(boost::bind(&Private::doQuit, this));
 		hupConnection = ProcessQuit::instance()->hup.connect(boost::bind(&App::Private::reload, this));
@@ -250,6 +424,7 @@ public:
 
 		QStringList services = settings.value("runner/services").toStringList();
 
+		int workerCount = settings.value("proxy/workers", 1).toInt();
 		QStringList condure_in_specs = settings.value("proxy/condure_in_specs").toStringList();
 		trimlist(&condure_in_specs);
 		QStringList condure_in_stream_specs = settings.value("proxy/condure_in_stream_specs").toStringList();
@@ -277,8 +452,10 @@ public:
 		QString handler_inspect_spec = settings.value("proxy/handler_inspect_spec").toString();
 		QString handler_accept_spec = settings.value("proxy/handler_accept_spec").toString();
 		QString handler_retry_in_spec = settings.value("proxy/handler_retry_in_spec").toString();
-		QString handler_ws_control_in_spec = settings.value("proxy/handler_ws_control_in_spec").toString();
-		QString handler_ws_control_out_spec = settings.value("proxy/handler_ws_control_out_spec").toString();
+		QStringList handler_ws_control_init_specs = settings.value("proxy/handler_ws_control_init_specs").toStringList();
+		trimlist(&handler_ws_control_init_specs);
+		QStringList handler_ws_control_stream_specs = settings.value("proxy/handler_ws_control_stream_specs").toStringList();
+		trimlist(&handler_ws_control_stream_specs);
 		QString stats_spec = settings.value("proxy/stats_spec").toString();
 		QString command_spec = settings.value("proxy/command_spec").toString();
 		QStringList intreq_in_specs = settings.value("proxy/intreq_in_specs").toStringList();
@@ -289,7 +466,7 @@ public:
 		trimlist(&intreq_out_specs);
 		bool ok;
 		int ipcFileMode = settings.value("proxy/ipc_file_mode", -1).toString().toInt(&ok, 8);
-		int maxWorkers = settings.value("proxy/max_open_requests", -1).toInt();
+		int sessionsMax = settings.value("proxy/max_open_requests", -1).toInt();
 		QString routesFile = settings.value("proxy/routesfile").toString();
 		bool debug = settings.value("proxy/debug").toBool();
 		bool autoCrossOrigin = settings.value("proxy/auto_cross_origin").toBool();
@@ -345,6 +522,23 @@ public:
 		if(updatesCheck == "true")
 			updatesCheck = "check";
 
+		// sessionsMax should not exceed clientMaxconn
+		if(sessionsMax >= 0)
+			sessionsMax = qMin(sessionsMax, clientMaxconn);
+		else
+			sessionsMax = clientMaxconn;
+
+		if(!args.routeLines.isEmpty())
+		{
+			domainMap = new DomainMap(this);
+			foreach(const QString &line, args.routeLines)
+				domainMap->addRouteLine(line);
+		}
+		else
+			domainMap = new DomainMap(routesFile, this);
+
+		changedConnection = domainMap->changed.connect(boost::bind(&Private::domainMap_changed, this));
+
 		Engine::Configuration config;
 		config.appVersion = Config::get().version;
 		config.clientId = "pushpin-proxy_" + QByteArray::number(QCoreApplication::applicationPid());
@@ -375,19 +569,15 @@ public:
 		config.inspectSpec = handler_inspect_spec;
 		config.acceptSpec = handler_accept_spec;
 		config.retryInSpec = handler_retry_in_spec;
-		config.wsControlInSpec = handler_ws_control_in_spec;
-		config.wsControlOutSpec = handler_ws_control_out_spec;
+		config.wsControlInitSpecs = handler_ws_control_init_specs;
+		config.wsControlStreamSpecs = handler_ws_control_stream_specs;
 		config.statsSpec = stats_spec;
 		config.commandSpec = command_spec;
 		config.intServerInSpecs = intreq_in_specs;
 		config.intServerInStreamSpecs = intreq_in_stream_specs;
 		config.intServerOutSpecs = intreq_out_specs;
 		config.ipcFileMode = ipcFileMode;
-		config.maxWorkers = maxWorkers;
-		if(!args.routeLines.isEmpty())
-			config.routeLines = args.routeLines;
-		else
-			config.routesFile = routesFile;
+		config.sessionsMax = sessionsMax / workerCount;
 		config.debug = debug;
 		config.autoCrossOrigin = autoCrossOrigin;
 		config.acceptXForwardedProto = acceptXForwardedProtocol;
@@ -407,7 +597,6 @@ public:
 		config.updatesCheck = updatesCheck;
 		config.organizationName = organizationName;
 		config.quietCheck = args.quietCheck;
-		config.connectionsMax = clientMaxconn;
 		config.statsConnectionSend = statsConnectionSend;
 		config.statsConnectionTtl = statsConnectionTtl;
 		config.statsConnectionsMaxTtl = statsConnectionsMaxTtl;
@@ -415,11 +604,43 @@ public:
 		config.prometheusPort = prometheusPort;
 		config.prometheusPrefix = prometheusPrefix;
 
-		engine = new Engine(this);
-		if(!engine->start(config))
+		for(int n = 0; n < workerCount; ++n)
 		{
-			q->quit(0);
-			return;
+			Engine::Configuration wconfig = config;
+
+			wconfig.id = n;
+
+			if(workerCount > 1)
+			{
+				wconfig.clientId += '-' + QByteArray::number(n);
+
+				wconfig.inspectSpec = suffixSpec(wconfig.inspectSpec, n);
+				wconfig.acceptSpec = suffixSpec(wconfig.acceptSpec, n);
+				wconfig.retryInSpec = suffixSpec(wconfig.retryInSpec, n);
+				wconfig.wsControlInitSpecs = suffixSpecs(wconfig.wsControlInitSpecs, n);
+				wconfig.wsControlStreamSpecs = suffixSpecs(wconfig.wsControlStreamSpecs, n);
+				wconfig.statsSpec = suffixSpec(wconfig.statsSpec, n);
+				wconfig.commandSpec = suffixSpec(wconfig.commandSpec, n);
+				wconfig.intServerInSpecs = suffixSpecs(wconfig.intServerInSpecs, n);
+				wconfig.intServerInStreamSpecs = suffixSpecs(wconfig.intServerInStreamSpecs, n);
+				wconfig.intServerOutSpecs = suffixSpecs(wconfig.intServerOutSpecs, n);
+			}
+
+			EngineThread *t = new EngineThread(wconfig, domainMap);
+			if(!t->start())
+			{
+				delete t;
+
+				for(EngineThread *t : threads)
+					delete t;
+
+				threads.clear();
+
+				q->quit(0);
+				return;
+			}
+
+			threads.push_back(t);
 		}
 
 		log_info("started");
@@ -430,7 +651,14 @@ private slots:
 	{
 		log_info("reloading");
 		log_rotate();
-		engine->reload();
+
+		domainMap->reload();
+	}
+
+	void domainMap_changed()
+	{
+		for(EngineThread *t : threads)
+			t->routesChanged();
 	}
 
 	void doQuit()
@@ -440,8 +668,13 @@ private slots:
 		// remove the handler, so if we get another signal then we crash out
 		ProcessQuit::cleanup();
 
-		delete engine;
-		engine = 0;
+		for(EngineThread *t : threads)
+			t->stop();
+
+		for(EngineThread *t : threads)
+			delete t;
+
+		threads.clear();
 
 		log_info("stopped");
 		q->quit(0);
