@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2012-2023 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -26,11 +26,13 @@
 #include <assert.h>
 #include "qzmqsocket.h"
 #include "qzmqvalve.h"
+#include "qzmqreqmessage.h"
 #include "tnetstring.h"
 #include "packet/httpresponsedata.h"
 #include "packet/retryrequestpacket.h"
 #include "packet/statspacket.h"
 #include "packet/zrpcrequestpacket.h"
+#include "qtcompat.h"
 #include "rtimer.h"
 #include "log.h"
 #include "inspectdata.h"
@@ -87,15 +89,28 @@ public:
 		}
 	};
 
+	struct RequestSessionConnections {
+		Connection inspectedConnection;
+		Connection inspectErrorConnection;
+		Connection finishedConnection;
+		Connection finishedByAcceptConnection;
+	};
+
+	struct ProxySessionConnections {
+		Connection addNotAllowedConnection;
+		Connection finishedConnection;
+		Connection reqSessionDestroyedConnection;
+	};
+
 	Engine *q;
 	bool destroying;
+	DomainMap *domainMap;
 	Configuration config;
 	ZhttpManager *zhttpIn;
 	ZhttpManager *intZhttpIn;
 	ZRoutes *zroutes;
 	ZrpcManager *inspect;
-	WsControlManager *wsControl;
-	DomainMap *domainMap;
+	std::unique_ptr<WsControlManager> wsControl;
 	ZrpcChecker *inspectChecker;
 	StatsManager *stats;
 	ZrpcManager *command;
@@ -110,23 +125,25 @@ public:
 	ConnectionManager connectionManager;
 	Updater *updater;
 	LogUtil::Config logConfig;
-	Connection changedConnection;
 	Connection cmdReqReadyConnection;
 	Connection sessionReadyConnection;
 	Connection requestReadyConnection;
 	Connection socketReadyConnection;
 	Connection iRequestReadyConnection;
-	
-	Private(Engine *_q) :
+	map<RequestSession*, RequestSessionConnections> reqSessionConnectionMap;
+	map<ProxySession*, ProxySessionConnections> proxySessionConnectionMap;
+	Connection connMaxConnection;
+	Connection rrConnection;
+
+	Private(Engine *_q, DomainMap *_domainMap) :
 		QObject(_q),
 		q(_q),
 		destroying(false),
+		domainMap(_domainMap),
 		zhttpIn(0),
 		intZhttpIn(0),
 		zroutes(0),
 		inspect(0),
-		wsControl(0),
-		domainMap(0),
 		inspectChecker(0),
 		stats(0),
 		command(0),
@@ -142,7 +159,8 @@ public:
 	{
 		destroying = true;
 
-		// need to delete all objects that may have outgoing connections before zroutes
+		// need to delete all objects that may have connections before
+		// deleting zhttpmanagers/zroutes
 
 		delete updater;
 
@@ -167,10 +185,16 @@ public:
 
 		wsProxyItemsBySession.clear();
 
-		foreach(RequestSession *rs, requestSessions)
+		foreach(RequestSession *rs, requestSessions){
+			reqSessionConnectionMap.erase(rs);
 			delete rs;
+		}
 		requestSessions.clear();
 
+		// may have background connections
+		delete sockJsManager;
+		sockJsManager = 0;
+
 		WebSocketOverHttp::clearDisconnectManager();
 
 		// need to make sure this is deleted before inspect manager
@@ -182,24 +206,13 @@ public:
 	{
 		config = _config;
 
-		// up to 10 timers per connection
-		RTimer::init(config.connectionsMax * 10);
+		// up to 10 timers per session
+		RTimer::init(config.sessionsMax * 10);
 
 		logConfig.fromAddress = config.logFrom;
 		logConfig.userAgent = config.logUserAgent;
 
-		WebSocketOverHttp::setMaxManagedDisconnects(config.maxWorkers);
-
-		if(!config.routeLines.isEmpty())
-		{
-			domainMap = new DomainMap(this);
-			foreach(const QString &line, config.routeLines)
-				domainMap->addRouteLine(line);
-		}
-		else
-			domainMap = new DomainMap(config.routesFile, this);
-
-		changedConnection = domainMap->changed.connect(boost::bind(&Private::domainMap_changed, this));
+		WebSocketOverHttp::setMaxManagedDisconnects(config.sessionsMax);
 
 		zhttpIn = new ZhttpManager(this);
 		requestReadyConnection = zhttpIn->requestReady.connect(boost::bind(&Private::zhttpIn_requestReady, this));
@@ -251,6 +264,7 @@ public:
 		if(!config.acceptSpec.isEmpty())
 		{
 			accept = new ZrpcManager(this);
+			accept->setInstanceId(config.clientId);
 			accept->setBind(true);
 			accept->setIpcFileMode(config.ipcFileMode);
 			if(!accept->setClientSpecs(QStringList() << config.acceptSpec))
@@ -265,8 +279,9 @@ public:
 
 		if(!config.retryInSpec.isEmpty())
 		{
-			handler_retry_in_sock = new QZmq::Socket(QZmq::Socket::Pull, this);
+			handler_retry_in_sock = new QZmq::Socket(QZmq::Socket::Router, this);
 
+			handler_retry_in_sock->setIdentity(config.clientId);
 			handler_retry_in_sock->setHwm(DEFAULT_HWM);
 
 			QString errorMessage;
@@ -277,36 +292,37 @@ public:
 			}
 
 			handler_retry_in_valve = new QZmq::Valve(handler_retry_in_sock, this);
-			connect(handler_retry_in_valve, &QZmq::Valve::readyRead, this, &Private::handler_retry_in_readyRead);
+			rrConnection = handler_retry_in_valve->readyRead.connect(boost::bind(&Private::handler_retry_in_readyRead, this, boost::placeholders::_1));
 		}
 
 		if(handler_retry_in_valve)
 			handler_retry_in_valve->open();
 
-		if(!config.wsControlInSpec.isEmpty() && !config.wsControlOutSpec.isEmpty())
+		if(!config.wsControlInitSpecs.isEmpty() && !config.wsControlStreamSpecs.isEmpty())
 		{
-			wsControl = new WsControlManager(this);
+			wsControl = std::make_unique<WsControlManager>();
 
+			wsControl->setIdentity(config.clientId);
 			wsControl->setIpcFileMode(config.ipcFileMode);
 
-			if(!wsControl->setInSpec(config.wsControlInSpec))
+			if(!wsControl->setInitSpecs(config.wsControlInitSpecs))
 			{
-				log_error("unable to bind to handler_ws_control_in_spec: %s", qPrintable(config.wsControlInSpec));
+				log_error("unable to bind to handler_ws_control_init_specs: %s", qPrintable(config.wsControlInitSpecs.join(", ")));
 				return false;
 			}
 
-			if(!wsControl->setOutSpec(config.wsControlOutSpec))
+			if(!wsControl->setStreamSpecs(config.wsControlStreamSpecs))
 			{
-				log_error("unable to bind to handler_ws_control_out_spec: %s", qPrintable(config.wsControlOutSpec));
+				log_error("unable to bind to handler_ws_control_stream_specs: %s", qPrintable(config.wsControlStreamSpecs.join(", ")));
 				return false;
 			}
 		}
 
 		if(!config.statsSpec.isEmpty() || !config.prometheusPort.isEmpty())
 		{
-			stats = new StatsManager(config.connectionsMax, 0, this);
+			stats = new StatsManager(config.sessionsMax, 0, this);
 
-			connect(stats, &StatsManager::connMax, this, &Private::stats_connMax);
+			connMaxConnection = stats->connMax.connect(boost::bind(&Private::stats_connMax, this, boost::placeholders::_1));
 
 			stats->setInstanceId(config.clientId);
 			stats->setIpcFileMode(config.ipcFileMode);
@@ -357,14 +373,15 @@ public:
 		}
 
 		// init zroutes
-		domainMap_changed();
+		routesChanged();
 
 		return true;
 	}
 
-	void reload()
+	void routesChanged()
 	{
-		domainMap->reload();
+		// connect to new zhttp targets, disconnect from old
+		zroutes->setup(domainMap->zhttpRoutes());
 	}
 
 	void doProxy(RequestSession *rs, const InspectData *idata = 0)
@@ -392,9 +409,11 @@ public:
 
 			ps = new ProxySession(zroutes, accept, logConfig, stats);
 			// TODO: use callbacks for performance
-			connect(ps, &ProxySession::addNotAllowed, this, &Private::ps_addNotAllowed);
-			connect(ps, &ProxySession::finished, this, &Private::ps_finished);
-			connect(ps, &ProxySession::requestSessionDestroyed, this, &Private::ps_requestSessionDestroyed);
+			proxySessionConnectionMap[ps] = {
+				ps->addNotAllowed.connect(boost::bind(&Private::ps_addNotAllowed, this, ps)),
+				ps->finished.connect(boost::bind(&Private::ps_finished, this, ps)),
+				ps->requestSessionDestroyed.connect(boost::bind(&Private::ps_requestSessionDestroyed, this, boost::placeholders::_1, boost::placeholders::_2))
+			};
 
 			ps->setRoute(route);
 			ps->setDefaultSigKey(config.sigIss, config.sigKey);
@@ -425,7 +444,7 @@ public:
 
 		// proxysession will take it from here
 		// TODO: use callbacks for performance
-		rs->disconnect(this);
+		reqSessionConnectionMap.erase(rs);
 
 		ps->add(rs);
 	}
@@ -434,7 +453,7 @@ public:
 	{
 		QByteArray cid = connectionManager.addConnection(sock);
 
-		WsProxySession *ps = new WsProxySession(zroutes, &connectionManager, logConfig, stats, wsControl);
+		WsProxySession *ps = new WsProxySession(zroutes, &connectionManager, logConfig, stats, wsControl.get());
 		ps->finishedByPassthroughCallback().add(Private::wsps_finishedByPassthrough_cb, this);
 
 		connectionManager.setProxyForConnection(sock, ps);
@@ -466,13 +485,13 @@ public:
 
 	bool canTake()
 	{
-		// don't accept new connections during shutdown
+		// don't accept new sessions during shutdown
 		if(destroying)
 			return false;
 
-		// don't accept new connections if we're servicing maximum
-		int curWorkers = requestSessions.count() + wsProxyItemsBySession.count();
-		if(config.maxWorkers != -1 && curWorkers >= config.maxWorkers)
+		// don't accept new sessions if we're servicing maximum
+		int curSessions = requestSessions.count() + wsProxyItemsBySession.count();
+		if(curSessions >= config.sessionsMax)
 			return false;
 
 		return true;
@@ -534,7 +553,8 @@ public:
 				routeId = QString::fromUtf8(req->requestHeaders().get("Pushpin-Route"));
 		}
 
-		RequestSession *rs;
+		RequestSession *rs = new RequestSession(config.id, domainMap, sockJsManager, inspect, inspectChecker, accept, stats);
+
 		if(passthroughData.isValid() && !preferInternal)
 		{
 			// passthrough request with preferInternal=false. in this case,
@@ -566,7 +586,6 @@ public:
 
 			route.targets += target;
 
-			rs = new RequestSession(stats);
 			rs->setRoute(route);
 		}
 		else
@@ -575,22 +594,29 @@ public:
 			// request with preferInternal=true. in that case, use domainmap
 			// for lookup, with route ID if available
 
-			rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, stats);
+			rs->setRouteId(routeId);
+		}
+
+		if(!passthroughData.isValid())
+		{
+			// these only make sense on regular requests
+
 			rs->setDebugEnabled(config.debug);
 			rs->setAutoCrossOrigin(config.autoCrossOrigin);
 			rs->setPrefetchSize(config.inspectPrefetch);
 			rs->setDefaultUpstreamKey(config.upstreamKey);
 			rs->setXffRules(config.xffUntrustedRule, config.xffTrustedRule);
-			rs->setRouteId(routeId);
 		}
 
 		rs->setAutoShare(autoShare);
 
 		// TODO: use callbacks for performance
-		connect(rs, &RequestSession::inspected, this, &Private::rs_inspected);
-		connect(rs, &RequestSession::inspectError, this, &Private::rs_inspectError);
-		connect(rs, &RequestSession::finished, this, &Private::rs_finished);
-		connect(rs, &RequestSession::finishedByAccept, this, &Private::rs_finishedByAccept);
+		reqSessionConnectionMap[rs] = {
+			rs->inspected.connect(boost::bind(&Private::rs_inspected, this, boost::placeholders::_1, rs)),
+			rs->inspectError.connect(boost::bind(&Private::rs_inspectError, this, rs)),
+			rs->finished.connect(boost::bind(&Private::rs_finished, this, rs)),
+			rs->finishedByAccept.connect(boost::bind(&Private::rs_finishedByAccept, this, rs))
+		};
 
 		requestSessions += rs;
 
@@ -611,7 +637,7 @@ public:
 
 		QUrl requestUri = sock->requestUri();
 
-		log_debug("IN ws id=%s, %s", sock->rid().second.data(), requestUri.toEncoded().data());
+		log_debug("worker %d: IN ws id=%s, %s", config.id, sock->rid().second.data(), requestUri.toEncoded().data());
 
 		bool isSecure = (requestUri.scheme() == "wss");
 		QString host = requestUri.host();
@@ -707,20 +733,25 @@ private:
 	{
 		tryTakeNext();
 	}
+
 	void intZhttpIn_requestReady()
 	{
 		tryTakeNext();
 	}
+
 	void sockjs_sessionReady()
 	{
 		tryTakeNext();
 	}
 
-private slots:
-	void rs_inspected(const InspectData &idata)
+	void rs_inspectError(RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
+		// default action is to proxy without sharing
+		doProxy(rs);
+	}
 
+	void rs_inspected(const InspectData &idata, RequestSession *rs)
+	{
 		// if we get here, then the request must be proxied. if it was to be directly
 		//   accepted, then finishedByAccept would have been emitted instead
 		assert(idata.doProxy);
@@ -728,43 +759,31 @@ private slots:
 		doProxy(rs, &idata);
 	}
 
-	void rs_inspectError()
+	void rs_finished(RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
-		// default action is to proxy without sharing
-		doProxy(rs);
-	}
-
-	void rs_finished()
-	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		if(!rs->isSockJs())
 			logFinished(rs);
 
 		requestSessions.remove(rs);
+		reqSessionConnectionMap.erase(rs);
 		delete rs;
 
 		tryTakeNext();
 	}
 
-	void rs_finishedByAccept()
+	void rs_finishedByAccept(RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		logFinished(rs, true);
 
 		requestSessions.remove(rs);
+		reqSessionConnectionMap.erase(rs);
 		delete rs;
 
 		tryTakeNext();
 	}
 
-	void ps_addNotAllowed()
+	void ps_addNotAllowed(ProxySession *ps)
 	{
-		ProxySession *ps = (ProxySession *)sender();
-
 		ProxyItem *i = proxyItemsBySession.value(ps);
 		assert(i);
 
@@ -776,13 +795,13 @@ private slots:
 		}
 	}
 
-	void ps_finished()
+	void ps_finished(ProxySession *ps)
 	{
-		ProxySession *ps = (ProxySession *)sender();
-
 		ProxyItem *i = proxyItemsBySession.value(ps);
 		assert(i);
 
+		proxySessionConnectionMap.erase(ps);
+		
 		if(i->shared)
 			proxyItemsByKey.remove(i->key);
 		proxyItemsBySession.remove(i->ps);
@@ -827,16 +846,19 @@ private slots:
 		tryTakeNext();
 	}
 
+private:
 	void handler_retry_in_readyRead(const QList<QByteArray> &message)
 	{
-		if(message.count() != 1)
+		QZmq::ReqMessage req(message);
+
+		if(req.content().count() != 1)
 		{
 			log_warning("retry: received message with parts != 1, skipping");
 			return;
 		}
 
 		bool ok;
-		QVariant data = TnetString::toVariant(message[0], 0, &ok);
+		QVariant data = TnetString::toVariant(req.content()[0], 0, &ok);
 		if(!ok)
 		{
 			log_warning("retry: received message with invalid format (tnetstring parse failed), skipping");
@@ -855,9 +877,6 @@ private slots:
 
 		log_debug("IN (retry) %s %s", qPrintable(p.requestData.method), p.requestData.uri.toEncoded().data());
 
-		if(p.retrySeq >= 0)
-			stats->setRetrySeq(p.route, p.retrySeq);
-
 		InspectData idata;
 		if(p.haveInspectInfo)
 		{
@@ -886,7 +905,7 @@ private slots:
 
 			ZhttpRequest *zhttpRequest = zhttpIn->createRequestFromState(ss);
 
-			RequestSession *rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, stats);
+			RequestSession *rs = new RequestSession(config.id, domainMap, sockJsManager, inspect, inspectChecker, accept, stats);
 
 			requestSessions += rs;
 
@@ -899,7 +918,7 @@ private slots:
 			// note: if the routing table was changed, there's a chance the request
 			//   might get a different route id this time around. this could confuse
 			//   stats processors tracking route+connection mappings.
-			rs->startRetry(zhttpRequest, req.debug, req.autoCrossOrigin, req.jsonpCallback, req.jsonpExtendedResponse, req.unreportedTime);
+			rs->startRetry(zhttpRequest, req.debug, req.autoCrossOrigin, req.jsonpCallback, req.jsonpExtendedResponse, req.unreportedTime, p.retrySeq);
 
 			doProxy(rs, p.haveInspectInfo ? &idata : 0);
 		}
@@ -917,7 +936,6 @@ private slots:
 		}
 	}
 
-private:
 	void command_requestReady()
 	{
 		ZrpcRequest *req = command->takeNext();
@@ -931,7 +949,7 @@ private:
 			}
 
 			QVariantHash args = req->args();
-			if(!args.contains("ids") || args["ids"].type() != QVariant::List)
+			if(!args.contains("ids") || typeId(args["ids"]) != QMetaType::QVariantList)
 			{
 				req->respondError("bad-format");
 				delete req;
@@ -944,7 +962,7 @@ private:
 			QList<QByteArray> ids;
 			foreach(const QVariant &vid, vids)
 			{
-				if(vid.type() != QVariant::ByteArray)
+				if(typeId(vid) != QMetaType::QByteArray)
 				{
 					ok = false;
 					break;
@@ -971,7 +989,7 @@ private:
 		else if(req->method() == "refresh")
 		{
 			QVariantHash args = req->args();
-			if(!args.contains("cid") || args["cid"].type() != QVariant::ByteArray)
+			if(!args.contains("cid") || typeId(args["cid"]) != QMetaType::QByteArray)
 			{
 				req->respondError("bad-format");
 				delete req;
@@ -997,7 +1015,7 @@ private:
 		else if(req->method() == "report")
 		{
 			QVariantHash args = req->args();
-			if(!args.contains("stats") || args["stats"].type() != QVariant::Hash)
+			if(!args.contains("stats") || typeId(args["stats"]) != QMetaType::QVariantHash)
 			{
 				req->respondError("bad-format");
 				delete req;
@@ -1047,18 +1065,12 @@ private:
 
 		delete req;
 	}
-
-	void domainMap_changed()
-	{
-		// connect to new zhttp targets, disconnect from old
-		zroutes->setup(domainMap->zhttpRoutes());
-	}
 };
 
-Engine::Engine(QObject *parent) :
+Engine::Engine(DomainMap *domainMap, QObject *parent) :
 	QObject(parent)
 {
-	d = new Private(this);
+	d = new Private(this, domainMap);
 }
 
 Engine::~Engine()
@@ -1076,9 +1088,9 @@ bool Engine::start(const Configuration &config)
 	return d->start(config);
 }
 
-void Engine::reload()
+void Engine::routesChanged()
 {
-	d->reload();
+	d->routesChanged();
 }
 
 #include "engine.moc"
