@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2012-2023 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -31,6 +31,7 @@
 #include "packet/statspacket.h"
 #include "packet/httprequestdata.h"
 #include "packet/httpresponsedata.h"
+#include "qtcompat.h"
 #include "bufferlist.h"
 #include "log.h"
 #include "jwt.h"
@@ -47,6 +48,8 @@
 #include "acceptrequest.h"
 #include "testhttprequest.h"
 
+using std::map;
+
 #define MAX_ACCEPT_REQUEST_BODY 100000
 
 // NOTE: if this value is ever changed, fix enginetest to match
@@ -102,6 +105,21 @@ public:
 		}
 	};
 
+	struct RequestSessionConnections {
+		Connection bytesWrittenConnection;
+		Connection errorRespondingConnection;
+		Connection pausedConnection;
+		Connection finishedConnection;
+		Connection headerBytesSentConnection;
+		Connection bodyBytesSentConnection;
+	};
+
+	struct ZhttpReqConnections {
+		Connection readyReadConnection;
+		Connection writeBytesChangedConnection;
+		Connection errorConnection;
+	};
+
 	ProxySession *q;
 	State state;
 	ZRoutes *zroutes;
@@ -151,10 +169,9 @@ public:
 	StatsManager *statsManager;
 	Connection inReqReadyReadConnection;
 	Connection inReqErrorConnection;
-	Connection readyReadConnection;
-	Connection writeBytesChangedConnection;
-	Connection errorConnection;
+	ZhttpReqConnections zhttpReqConnections;
 	Connection finishedConnection;
+	map<RequestSession*, RequestSessionConnections> reqSessionConnectionMap;
 
 	Private(ProxySession *_q, ZRoutes *_zroutes, ZrpcManager *_acceptManager, const LogUtil::Config &_logConfig, StatsManager *_statsManager) :
 		QObject(_q),
@@ -198,7 +215,7 @@ public:
 		foreach(SessionItem *si, sessionItems)
 		{
 			// emitting a signal here is gross, but this way the engine cleans up the request sessions
-			emit q->requestSessionDestroyed(si->rs, false);
+			q->requestSessionDestroyed(si->rs, false);
 			delete si->rs;
 			delete si;
 		}
@@ -238,12 +255,14 @@ public:
 
 		sessionItems += si;
 		sessionItemsBySession.insert(rs, si);
-		connect(rs, &RequestSession::bytesWritten, this, &Private::rs_bytesWritten);
-		connect(rs, &RequestSession::errorResponding, this, &Private::rs_errorResponding);
-		connect(rs, &RequestSession::finished, this, &Private::rs_finished);
-		connect(rs, &RequestSession::paused, this, &Private::rs_paused);
-		connect(rs, &RequestSession::headerBytesSent, this, &Private::rs_headerBytesSent);
-		connect(rs, &RequestSession::bodyBytesSent, this, &Private::rs_bodyBytesSent);
+		reqSessionConnectionMap[rs] = {
+			rs->bytesWritten.connect(boost::bind(&Private::rs_bytesWritten, this, boost::placeholders::_1, rs)),
+			rs->errorResponding.connect(boost::bind(&Private::rs_errorResponding, this, rs)),
+			rs->paused.connect(boost::bind(&Private::rs_paused, this, rs)),
+			rs->finished.connect(boost::bind(&Private::rs_finished, this, rs)),
+			rs->headerBytesSent.connect(boost::bind(&Private::rs_headerBytesSent, this, boost::placeholders::_1, rs)),
+			rs->bodyBytesSent.connect(boost::bind(&Private::rs_bodyBytesSent, this, boost::placeholders::_1, rs))
+		};
 
 		HttpRequestData rsRequestData = rs->requestData();
 
@@ -442,9 +461,11 @@ public:
 			zhttpRequest->setParent(this);
 		}
 
-		readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
-		writeBytesChangedConnection = zhttpRequest->writeBytesChanged.connect(boost::bind(&Private::zhttpRequest_writeBytesChanged, this));
-		errorConnection = zhttpRequest->error.connect(boost::bind(&Private::zhttpRequest_error, this));
+		zhttpReqConnections = {
+			zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this)),
+			zhttpRequest->writeBytesChanged.connect(boost::bind(&Private::zhttpRequest_writeBytesChanged, this)),
+			zhttpRequest->error.connect(boost::bind(&Private::zhttpRequest_error, this))
+		};
 
 		if(target.trusted)
 			zhttpRequest->setIgnorePolicies(true);
@@ -481,7 +502,8 @@ public:
 			// no need to track the primary request anymore
 			if(inRequest)
 			{
-				inRequest->request()->disconnect(this);
+				inReqReadyReadConnection.disconnect();
+				inReqErrorConnection.disconnect();
 				inRequest = 0;
 			}
 
@@ -533,7 +555,8 @@ public:
 		if(!requestBodySent && inRequest->request()->isInputFinished() && inRequest->request()->bytesAvailable() == 0)
 		{
 			// no need to track the primary request anymore
-			inRequest->request()->disconnect(this);
+			inReqReadyReadConnection.disconnect();
+			inReqErrorConnection.disconnect();
 			inRequest = 0;
 
 			requestBodySent = true;
@@ -603,6 +626,7 @@ public:
 
 	void rejectAll(int code, const QString &reason, const QString &errorMessage, const QString &debugErrorMessage)
 	{
+		zhttpReqConnections = ZhttpReqConnections();
 		// kill the active target request, if any
 		delete zhttpRequest;
 		zhttpRequest = 0;
@@ -890,7 +914,7 @@ public:
 
 			if(wasAllowed && !addAllowed)
 			{
-				emit q->addNotAllowed();
+				q->addNotAllowed();
 				if(!self)
 					return;
 			}
@@ -914,6 +938,7 @@ public:
 				return;
 			}
 
+			zhttpReqConnections = ZhttpReqConnections();			
 			delete zhttpRequest;
 			zhttpRequest = 0;
 
@@ -921,7 +946,7 @@ public:
 			if(addAllowed)
 			{
 				addAllowed = false;
-				emit q->addNotAllowed();
+				q->addNotAllowed();
 				if(!self)
 					return;
 			}
@@ -1167,11 +1192,8 @@ public:
 		}
 	}
 
-public slots:
-	void rs_bytesWritten(int count)
+	void rs_bytesWritten(int count, RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		log_debug("proxysession: %p response bytes written id=%s: %d", q, rs->rid().second.data(), count);
 
 		SessionItem *si = sessionItemsBySession.value(rs);
@@ -1187,10 +1209,8 @@ public slots:
 			tryResponseRead();
 	}
 
-	void rs_finished()
+	void rs_finished(RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		log_debug("proxysession: %p response finished id=%s", q, rs->rid().second.data());
 
 		SessionItem *si = sessionItemsBySession.value(rs);
@@ -1200,7 +1220,7 @@ public slots:
 			logFinished(si);
 
 		QPointer<QObject> self = this;
-		emit q->requestSessionDestroyed(si->rs, false);
+		q->requestSessionDestroyed(si->rs, false);
 		if(!self)
 			return;
 
@@ -1209,6 +1229,7 @@ public slots:
 
 		sessionItemsBySession.remove(rs);
 		sessionItems.remove(si);
+		reqSessionConnectionMap.erase(rs);
 		delete rs;
 
 		delete si;
@@ -1216,7 +1237,7 @@ public slots:
 		if(sessionItems.isEmpty())
 		{
 			log_debug("proxysession: %p finished by passthrough", q);
-			emit q->finished();
+			q->finished();
 		}
 		else if(wasInputRequest)
 		{
@@ -1228,10 +1249,8 @@ public slots:
 		}
 	}
 
-	void rs_paused()
+	void rs_paused(RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		log_debug("proxysession: %p response paused id=%s", q, rs->rid().second.data());
 
 		SessionItem *si = sessionItemsBySession.value(rs);
@@ -1321,10 +1340,10 @@ public slots:
 			if(!statsManager->connectionSendEnabled())
 			{
 				// flush max. the count will include the connections we just unregistered
-				adata.connMaxPackets += statsManager->getConnMaxPacket(route.id).toVariant();
+				adata.connMaxPackets += statsManager->getConnMaxPacket(route.statsRoute()).toVariant();
 
 				// flush max again to get the count without the connections
-				adata.connMaxPackets += statsManager->getConnMaxPacket(route.id).toVariant();
+				adata.connMaxPackets += statsManager->getConnMaxPacket(route.statsRoute()).toVariant();
 			}
 
 			acceptRequest = new AcceptRequest(acceptManager, this);
@@ -1333,10 +1352,8 @@ public slots:
 		}
 	}
 
-	void rs_errorResponding()
+	void rs_errorResponding(RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		log_debug("proxysession: %p response error id=%s", q, rs->rid().second.data());
 
 		SessionItem *si = sessionItemsBySession.value(rs);
@@ -1351,10 +1368,8 @@ public slots:
 		// don't destroy the RequestSession here. a finished signal will arrive next.
 	}
 
-	void rs_headerBytesSent(int count)
+	void rs_headerBytesSent(int count, RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		SessionItem *si = sessionItemsBySession.value(rs);
 		assert(si);
 
@@ -1362,10 +1377,8 @@ public slots:
 			incCounter(Stats::ClientHeaderBytesSent, count);
 	}
 
-	void rs_bodyBytesSent(int count)
+	void rs_bodyBytesSent(int count, RequestSession *rs)
 	{
-		RequestSession *rs = (RequestSession *)sender();
-
 		SessionItem *si = sessionItemsBySession.value(rs);
 		assert(si);
 
@@ -1373,13 +1386,13 @@ public slots:
 			incCounter(Stats::ClientContentBytesSent, count);
 	}
 
-public:
 	void acceptRequest_finished()
 	{
 		if(acceptRequest->success())
 		{
 			AcceptRequest::ResponseData rdata = acceptRequest->result();
 
+			finishedConnection.disconnect();
 			delete acceptRequest;
 			acceptRequest = 0;
 
@@ -1403,7 +1416,8 @@ public:
 				QPointer<QObject> self = this;
 				foreach(RequestSession *rs, toDestroy)
 				{
-					emit q->requestSessionDestroyed(rs, true);
+					q->requestSessionDestroyed(rs, true);
+					reqSessionConnectionMap.erase(rs);
 					delete rs;
 					if(!self)
 						return;
@@ -1411,7 +1425,7 @@ public:
 
 				log_debug("proxysession: %p finished for accept", q);
 				cleanup();
-				emit q->finished();
+				q->finished();
 			}
 			else
 			{
@@ -1454,7 +1468,7 @@ public:
 		{
 			// wake up receivers and reject
 
-			if(acceptRequest->errorCondition() == ZrpcRequest::ErrorFormat && ((ZrpcRequest *)acceptRequest)->result().type() == QVariant::ByteArray)
+			if(acceptRequest->errorCondition() == ZrpcRequest::ErrorFormat && typeId(((ZrpcRequest *)acceptRequest)->result()) == QMetaType::QByteArray)
 			{
 				QString errorString = QString::fromUtf8(((ZrpcRequest *)acceptRequest)->result().toByteArray());
 				QString msg = "Error while proxying to origin.";
@@ -1467,6 +1481,7 @@ public:
 				cannotAcceptAll();
 			}
 
+			finishedConnection.disconnect();
 			delete acceptRequest;
 			acceptRequest = 0;
 		}
