@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2012-2023 Fanout, Inc.
+ * Copyright (C) 2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -32,6 +33,7 @@
 #include <QJsonArray>
 #include "packet/httprequestdata.h"
 #include "packet/httpresponsedata.h"
+#include "qtcompat.h"
 #include "bufferlist.h"
 #include "log.h"
 #include "layertracker.h"
@@ -145,7 +147,15 @@ public:
 		RespondingInternal
 	};
 
+	struct ZhttpReqConnections {
+		Connection readyReadConnection;
+		Connection pausedConnection;
+		Connection errorConnection;
+		Connection bytesWrittenConnection;
+	};
+
 	RequestSession *q;
+	int workerId;
 	State state;
 	ZhttpRequest::Rid rid;
 	DomainMap *domainMap;
@@ -187,16 +197,14 @@ public:
 	XffRule xffRule;
 	XffRule xffTrustedRule;
 	bool isSockJs;
-	Connection errorConnection;
-	Connection pausedConnection;
-	Connection readyReadConnection;
-	Connection bytesWrittenConnection;
+	ZhttpReqConnections zhttpReqConnections;
 	Connection inspectFinishedConnection;
 	Connection acceptFinishedConnection;
 
-	Private(RequestSession *_q, DomainMap *_domainMap = 0, SockJsManager *_sockJsManager = 0, ZrpcManager *_inspectManager = 0, ZrpcChecker *_inspectChecker = 0, ZrpcManager *_acceptManager = 0, StatsManager *_stats = 0) :
+	Private(RequestSession *_q, int _workerId, DomainMap *_domainMap = 0, SockJsManager *_sockJsManager = 0, ZrpcManager *_inspectManager = 0, ZrpcChecker *_inspectChecker = 0, ZrpcManager *_acceptManager = 0, StatsManager *_stats = 0) :
 		QObject(_q),
 		q(_q),
+		workerId(_workerId),
 		state(Stopped),
 		domainMap(_domainMap),
 		sockJsManager(_sockJsManager),
@@ -235,13 +243,14 @@ public:
 	{
 		if(zhttpRequest)
 		{
+			zhttpReqConnections = ZhttpReqConnections();
 			delete zhttpRequest;
 			zhttpRequest = 0;
 		}
 
 		if(inspectRequest)
 		{
-			inspectRequest->disconnect(this);
+			inspectFinishedConnection.disconnect();
 			inspectChecker->give(inspectRequest);
 			inspectRequest = 0;
 		}
@@ -279,7 +288,7 @@ public:
 		peerAddress = req->peerAddress();
 		logicalPeerAddress = ProxyUtil::getLogicalAddress(requestData.headers, trusted ? xffTrustedRule : xffRule, peerAddress);
 
-		log_debug("IN id=%s, %s %s", rid.second.data(), qPrintable(requestData.method), requestData.uri.toEncoded().data());
+		log_debug("worker %d: IN id=%s, %s %s", workerId, rid.second.data(), qPrintable(requestData.method), requestData.uri.toEncoded().data());
 
 		bool isHttps = (requestData.uri.scheme() == "https");
 		QString host = requestData.uri.host();
@@ -300,13 +309,13 @@ public:
 				isSockJs = true;
 				sockJsManager->giveRequest(zhttpRequest, route.sockJsPath.length(), route.sockJsAsPath, route);
 				zhttpRequest = 0;
-				QMetaObject::invokeMethod(q, "finished", Qt::QueuedConnection);
+				QMetaObject::invokeMethod(this, "doFinished", Qt::QueuedConnection);
 				return;
 			}
 		}
 
-		pausedConnection = zhttpRequest->paused.connect(boost::bind(&Private::zhttpRequest_paused, this));
-		errorConnection = zhttpRequest->error.connect(boost::bind(&Private::zhttpRequest_error, this));
+		zhttpReqConnections.pausedConnection = zhttpRequest->paused.connect(boost::bind(&Private::zhttpRequest_paused, this));
+		zhttpReqConnections.errorConnection = zhttpRequest->error.connect(boost::bind(&Private::zhttpRequest_error, this));
 
 		if(!route.isNull())
 		{
@@ -379,19 +388,19 @@ public:
 
 		state = Prefetching;
 
-		readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
+		zhttpReqConnections.readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
 		processIncomingRequest();
 	}
 
-	void startRetry(int unreportedTime)
+	void startRetry(int unreportedTime, int retrySeq)
 	{
 		trusted = ProxyUtil::checkTrustedClient("requestsession", q, requestData, defaultUpstreamKey);
 
 		peerAddress = zhttpRequest->peerAddress();
 		logicalPeerAddress = ProxyUtil::getLogicalAddress(requestData.headers, trusted ? xffTrustedRule : xffRule, peerAddress);
 
-		pausedConnection = zhttpRequest->paused.connect(boost::bind(&Private::zhttpRequest_paused, this));
-		errorConnection = zhttpRequest->error.connect(boost::bind(&Private::zhttpRequest_error, this));
+		zhttpReqConnections.pausedConnection = zhttpRequest->paused.connect(boost::bind(&Private::zhttpRequest_paused, this));
+		zhttpReqConnections.errorConnection = zhttpRequest->error.connect(boost::bind(&Private::zhttpRequest_error, this));
 
 		state = WaitingForResponse;
 
@@ -417,6 +426,9 @@ public:
 
 		if(stats)
 		{
+			if(retrySeq >= 0)
+				stats->setRetrySeq(route.statsRoute(), retrySeq);
+
 			connectionRegistered = true;
 
 			int reportOffset = stats->connectionSendEnabled() ? -1 : qMax(unreportedTime, 0);
@@ -439,7 +451,7 @@ public:
 			{
 				// we've read enough body to start inspection
 
-				readyReadConnection.disconnect();
+				zhttpReqConnections.readyReadConnection.disconnect();
 
 				state = Inspecting;
 				requestData.body = in.toByteArray();
@@ -484,11 +496,11 @@ public:
 				//   disallow sharing before passing to proxysession. at that
 				//   point, proxysession will read the remainder of the data
 
-				readyReadConnection.disconnect();
+				zhttpReqConnections.readyReadConnection.disconnect();
 
 				state = WaitingForResponse;
 				requestData.body = in.take();
-				emit q->inspected(idata);
+				q->inspected(idata);
 			}
 		}
 		else if(state == ReceivingForAccept)
@@ -686,7 +698,7 @@ public:
 			{
 				vit.next();
 
-				if(vit.value().type() != QVariant::String)
+				if(typeId(vit.value()) != QMetaType::QString)
 				{
 					log_debug("requestsession: id=%s invalid _headers parameter, rejecting", rid.second.data());
 					*ok = false;
@@ -817,7 +829,7 @@ public:
 		if(zhttpRequest->isFinished())
 		{
 			cleanup();
-			emit q->finished();
+			q->finished();
 		}
 	}
 
@@ -867,14 +879,14 @@ public:
 	{
 		log_debug("requestsession: request error id=%s", rid.second.data());
 		cleanup();
-		emit q->finished();
+		q->finished();
 	}
 
 	void inspectRequest_finished()
 	{
 		if(!inspectRequest->success())
 		{
-			inspectRequest->disconnect(this);
+			inspectFinishedConnection.disconnect();
 			inspectChecker->give(inspectRequest);
 			inspectRequest = 0;
 
@@ -884,7 +896,7 @@ public:
 
 		idata = inspectRequest->result();
 
-		inspectRequest->disconnect(this);
+		inspectFinishedConnection.disconnect();
 		inspectChecker->give(inspectRequest);
 		inspectRequest = 0;
 
@@ -894,7 +906,7 @@ public:
 
 			// successful inspect indicated we should not proxy. in that case,
 			//   collect the body and accept
-			readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
+			zhttpReqConnections.readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
 			processIncomingRequest();
 		}
 		else
@@ -905,14 +917,14 @@ public:
 				//   request body, so let's try to read it now
 				state = Receiving;
 
-				readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
+				zhttpReqConnections.readyReadConnection = zhttpRequest->readyRead.connect(boost::bind(&Private::zhttpRequest_readyRead, this));
 				processIncomingRequest();
 			}
 			else
 			{
 				state = WaitingForResponse;
 				requestData.body = in.take();
-				emit q->inspected(idata);
+				q->inspected(idata);
 			}
 		}
 	}
@@ -923,6 +935,7 @@ public:
 		{
 			AcceptRequest::ResponseData rdata = acceptRequest->result();
 
+			acceptFinishedConnection.disconnect();
 			delete acceptRequest;
 			acceptRequest = 0;
 
@@ -931,11 +944,12 @@ public:
 				accepted = true;
 
 				// the request was paused, so deleting it will leave the peer session active
+				zhttpReqConnections = ZhttpReqConnections();
 				delete zhttpRequest;
 				zhttpRequest = 0;
 
 				cleanup();
-				emit q->finishedByAccept();
+				q->finishedByAccept();
 			}
 			else
 			{
@@ -953,6 +967,7 @@ public:
 		}
 		else
 		{
+			acceptFinishedConnection.disconnect();
 			delete acceptRequest;
 			acceptRequest = 0;
 
@@ -1004,7 +1019,7 @@ public slots:
 						zhttpRequest->writeBody(body);
 						responseBodySize += body.size();
 						zhttpRequest->endBody();
-						emit q->errorResponding();
+						q->errorResponding();
 						return;
 					}
 
@@ -1026,7 +1041,7 @@ public slots:
 						}
 					}
 
-					bytesWrittenConnection = zhttpRequest->bytesWritten.connect(boost::bind(&Private::zhttpRequest_bytesWritten, this, boost::placeholders::_1));
+					zhttpReqConnections.bytesWrittenConnection = zhttpRequest->bytesWritten.connect(boost::bind(&Private::zhttpRequest_bytesWritten, this, boost::placeholders::_1));
 
 					zhttpRequest->beginResponse(200, "OK", headers);
 
@@ -1051,14 +1066,14 @@ public slots:
 					zhttpRequest->writeBody(body);
 					responseBodySize += body.size();
 					zhttpRequest->endBody();
-					emit q->errorResponding();
+					q->errorResponding();
 					return;
 				}
 
 				headers += HttpHeader("Content-Type", "application/javascript");
 				headers += HttpHeader("Transfer-Encoding", "chunked");
 
-				bytesWrittenConnection = zhttpRequest->bytesWritten.connect(boost::bind(&Private::zhttpRequest_bytesWritten, this, boost::placeholders::_1));
+				zhttpReqConnections.bytesWrittenConnection = zhttpRequest->bytesWritten.connect(boost::bind(&Private::zhttpRequest_bytesWritten, this, boost::placeholders::_1));
 
 				zhttpRequest->beginResponse(200, "OK", headers);
 
@@ -1072,7 +1087,7 @@ public slots:
 				if(autoCrossOrigin)
 					Cors::applyCorsHeaders(requestData.headers, &responseData.headers);
 
-				bytesWrittenConnection = zhttpRequest->bytesWritten.connect(boost::bind(&Private::zhttpRequest_bytesWritten, this, boost::placeholders::_1));
+				zhttpReqConnections.bytesWrittenConnection = zhttpRequest->bytesWritten.connect(boost::bind(&Private::zhttpRequest_bytesWritten, this, boost::placeholders::_1));
 
 				zhttpRequest->beginResponse(responseData.code, responseData.reason, responseData.headers);
 			}
@@ -1119,7 +1134,7 @@ public slots:
 
 					// if we error while streaming, all we can do is give up
 					zhttpRequest->endBody();
-					emit q->errorResponding();
+					q->errorResponding();
 					return;
 				}
 
@@ -1166,20 +1181,19 @@ public slots:
 	void doInspectError()
 	{
 		state = WaitingForResponse;
-		emit q->inspectError();
+		q->inspectError();
+	}
+
+	void doFinished()
+	{
+		q->finished();
 	}
 };
 
-RequestSession::RequestSession(StatsManager *stats, QObject *parent) :
+RequestSession::RequestSession(int workerId, DomainMap *domainMap, SockJsManager *sockJsManager, ZrpcManager *inspectManager, ZrpcChecker *inspectChecker, ZrpcManager *acceptManager, StatsManager *stats, QObject *parent) :
 	QObject(parent)
 {
-	d = new Private(this, 0, 0, 0, 0, 0, stats);
-}
-
-RequestSession::RequestSession(DomainMap *domainMap, SockJsManager *sockJsManager, ZrpcManager *inspectManager, ZrpcChecker *inspectChecker, ZrpcManager *acceptManager, StatsManager *stats, QObject *parent) :
-	QObject(parent)
-{
-	d = new Private(this, domainMap, sockJsManager, inspectManager, inspectChecker, acceptManager, stats);
+	d = new Private(this, workerId, domainMap, sockJsManager, inspectManager, inspectChecker, acceptManager, stats);
 }
 
 RequestSession::~RequestSession()
@@ -1323,7 +1337,7 @@ void RequestSession::start(ZhttpRequest *req)
 	d->start(req);
 }
 
-void RequestSession::startRetry(ZhttpRequest *req, bool debug, bool autoCrossOrigin, const QByteArray &jsonpCallback, bool jsonpExtendedResponse, int unreportedTime)
+void RequestSession::startRetry(ZhttpRequest *req, bool debug, bool autoCrossOrigin, const QByteArray &jsonpCallback, bool jsonpExtendedResponse, int unreportedTime, int retrySeq)
 {
 	d->isRetry = true;
 	d->zhttpRequest = req;
@@ -1337,7 +1351,7 @@ void RequestSession::startRetry(ZhttpRequest *req, bool debug, bool autoCrossOri
 	d->requestData.headers = req->requestHeaders();
 	d->requestData.body = req->readBody();
 
-	d->startRetry(unreportedTime);
+	d->startRetry(unreportedTime, retrySeq);
 }
 
 void RequestSession::pause()
@@ -1357,7 +1371,7 @@ void RequestSession::startResponse(int code, const QByteArray &reason, const Htt
 {
 	assert(d->state == Private::ReceivingForAccept || d->state == Private::WaitingForResponse);
 
-	emit headerBytesSent(ZhttpManager::estimateResponseHeaderBytes(code, reason, headers));
+	headerBytesSent(ZhttpManager::estimateResponseHeaderBytes(code, reason, headers));
 
 	d->state = Private::RespondingStart;
 	d->responseData.code = code;
@@ -1372,7 +1386,7 @@ void RequestSession::writeResponseBody(const QByteArray &body)
 	assert(d->state == Private::RespondingStart || d->state == Private::Responding);
 	assert(!d->responseBodyFinished);
 
-	emit bodyBytesSent(body.size());
+	bodyBytesSent(body.size());
 
 	d->out += body;
 	d->responseUpdate();
