@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2015-2021 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -30,6 +30,7 @@
 #include <QJsonDocument>
 #include <QJsonObject>
 #include <QJsonArray>
+#include "qtcompat.h"
 #include "log.h"
 #include "bufferlist.h"
 #include "packet/httprequestdata.h"
@@ -37,6 +38,8 @@
 #include "zwebsocket.h"
 #include "sockjsmanager.h"
 
+using std::map;
+
 #define BUFFER_SIZE 200000
 #define KEEPALIVE_TIMEOUT 25
 #define UNCONNECTED_TIMEOUT 5
@@ -113,6 +116,20 @@ public:
 		}
 	};
 
+	struct WSConnections {
+		Connection readyReadConnection;
+		Connection framesWrittenConnection;
+		Connection writeBytesChangedConnection;
+		Connection closedConnection;
+		Connection peerClosedConnection;
+		Connection sockErrorConnection;
+	};
+
+	struct ReqConnections {
+		Connection bytesWrittenConnection;
+		Connection errorConnection;
+	};
+
 	SockJsSession *q;
 	SockJsManager *manager;
 	Mode mode;
@@ -146,8 +163,8 @@ public:
 	int peerCloseCode;
 	QString peerCloseReason;
 	bool updating;
-	Connection bytesWrittenConnection;
-	Connection errorConnection;
+	map<ZhttpRequest*, ReqConnections> reqConnectionMap;
+	WSConnections wsConnection;
 
 	Private(SockJsSession *_q) :
 		QObject(_q),
@@ -184,6 +201,7 @@ public:
 
 	void removeRequestItem(RequestItem *ri)
 	{
+		reqConnectionMap.erase(ri->req);
 		requests.remove(ri->req);
 		delete ri;
 	}
@@ -213,7 +231,8 @@ public:
 			assert(ri);
 
 			// detach req from RequestItem
-			requests.remove(req);
+			reqConnectionMap.erase(ri->req);
+			requests.remove(ri->req);
 			ri->req = 0;
 			delete ri;
 
@@ -231,6 +250,7 @@ public:
 		}
 		requests.clear();
 
+		wsConnection = WSConnections();
 		delete sock;
 		sock = 0;
 
@@ -256,17 +276,21 @@ public:
 
 			requests.insert(req, new RequestItem(req, jsonpCallback, RequestItem::Connect));
 
-			bytesWrittenConnection = req->bytesWritten.connect(boost::bind(&Private::req_bytesWritten, this, boost::placeholders::_1));
-			errorConnection = req->error.connect(boost::bind(&Private::req_error, this));
+			reqConnectionMap[req] = {
+				req->bytesWritten.connect(boost::bind(&Private::req_bytesWritten, this, boost::placeholders::_1, req)),
+				req->error.connect(boost::bind(&Private::req_error, this, req))
+			};
 		}
 		else
 		{
-			connect(sock, &ZWebSocket::readyRead, this, &Private::sock_readyRead);
-			connect(sock, &ZWebSocket::framesWritten, this, &Private::sock_framesWritten);
-			connect(sock, &ZWebSocket::writeBytesChanged, this, &Private::sock_writeBytesChanged);
-			connect(sock, &ZWebSocket::closed, this, &Private::sock_closed);
-			connect(sock, &ZWebSocket::peerClosed, this, &Private::sock_peerClosed);
-			connect(sock, &ZWebSocket::error, this, &Private::sock_error);
+			wsConnection = WSConnections{
+				sock->readyRead.connect(boost::bind(&Private::sock_readyRead, this)),
+				sock->framesWritten.connect(boost::bind(&Private::sock_framesWritten, this, boost::placeholders::_1, boost::placeholders::_2)),
+				sock->writeBytesChanged.connect(boost::bind(&Private::sock_writeBytesChanged, this)),
+				sock->closed.connect(boost::bind(&Private::sock_closed, this)),
+				sock->peerClosed.connect(boost::bind(&Private::sock_peerClosed, this)),
+				sock->error.connect(boost::bind(&Private::sock_error, this))
+			};
 		}
 	}
 
@@ -297,8 +321,10 @@ public:
 
 	void handleRequest(ZhttpRequest *_req, const QByteArray &jsonpCallback, const QByteArray &lastPart, const QByteArray &body)
 	{
-		bytesWrittenConnection = _req->bytesWritten.connect(boost::bind(&Private::req_bytesWritten, this, boost::placeholders::_1));
-		errorConnection = _req->error.connect(boost::bind(&Private::req_error, this));
+		reqConnectionMap[_req] = {
+			_req->bytesWritten.connect(boost::bind(&Private::req_bytesWritten, this, boost::placeholders::_1, _req)),
+			_req->error.connect(boost::bind(&Private::req_error, this, _req))
+		};
 
 		if(lastPart == "xhr" || lastPart == "jsonp")
 		{
@@ -357,6 +383,7 @@ public:
 						int at = kv.indexOf('=');
 						if(at == -1)
 							continue;
+
 						if(QUrl::fromPercentEncoding(kv.mid(0, at)) == "d")
 						{
 							param = QUrl::fromPercentEncoding(kv.mid(at + 1)).toUtf8();
@@ -386,7 +413,7 @@ public:
 			int bytes = 0;
 			foreach(const QVariant &vmessage, messages)
 			{
-				if(vmessage.type() != QVariant::String)
+				if(typeId(vmessage) != QMetaType::QString)
 				{
 					requests.insert(_req, new RequestItem(_req, jsonpCallback, RequestItem::Background, true));
 					respondError(_req, 400, "Bad Request", "Payload expected");
@@ -419,6 +446,11 @@ public:
 
 			tryRead();
 		}
+		else
+		{
+			requests.insert(_req, new RequestItem(_req, jsonpCallback, RequestItem::Background, true));
+			respondError(_req, 404, "Not Found", "Not Found");
+		}
 	}
 
 	void accept(const QByteArray &reason, const HttpHeaders &headers)
@@ -560,7 +592,7 @@ public:
 				state = Idle;
 				applyLinger();
 				cleanup();
-				QMetaObject::invokeMethod(q, "closed", Qt::QueuedConnection);
+				QMetaObject::invokeMethod(this, "doClosed", Qt::QueuedConnection);
 			}
 			else
 				tryWrite();
@@ -622,7 +654,7 @@ public:
 		if(bytes > 0)
 		{
 			QPointer<QObject> self = this;
-			emit q->writeBytesChanged();
+			q->writeBytesChanged();
 			if(!self)
 				return;
 		}
@@ -692,7 +724,7 @@ public:
 
 			if(emitReadyRead)
 			{
-				emit q->readyRead();
+				q->readyRead();
 				if(!self)
 					return false;
 			}
@@ -767,7 +799,7 @@ public:
 				int bytes = 0;
 				foreach(const QVariant &vmessage, messages)
 				{
-					if(vmessage.type() != QVariant::String)
+					if(typeId(vmessage) != QMetaType::QString)
 					{
 						error = true;
 						break;
@@ -799,7 +831,7 @@ public:
 			{
 				state = Idle;
 				cleanup();
-				emit q->error();
+				q->error();
 
 				// stop signals
 				return false;
@@ -807,7 +839,7 @@ public:
 
 			if(emitReadyRead)
 			{
-				emit q->readyRead();
+				q->readyRead();
 				if(!self)
 					return false;
 			}
@@ -850,7 +882,7 @@ public:
 			pendingWrittenBytes = 0;
 		}
 
-		emit q->framesWritten(count, contentBytes);
+		q->framesWritten(count, contentBytes);
 	}
 
 	QVariant applyLinger()
@@ -871,12 +903,10 @@ public:
 		return closeValue;
 	}
 
-private slots:
-	void req_bytesWritten(int count)
+	void req_bytesWritten(int count, ZhttpRequest *_req)
 	{
 		Q_UNUSED(count);
 
-		ZhttpRequest *_req = (ZhttpRequest *)sender();
 		RequestItem *ri = requests.value(_req);
 		assert(ri);
 
@@ -903,7 +933,7 @@ private slots:
 					state = Idle;
 					removeRequestItem(ri);
 					cleanup();
-					emit q->closed();
+					q->closed();
 					return;
 				}
 				else if(ri->type == RequestItem::Receive)
@@ -920,7 +950,7 @@ private slots:
 					state = Idle;
 					removeRequestItem(ri);
 					cleanup();
-					emit q->closed();
+					q->closed();
 					return;
 				}
 			}
@@ -929,9 +959,8 @@ private slots:
 		}
 	}
 
-	void req_error()
+	void req_error(ZhttpRequest *_req)
 	{
-		ZhttpRequest *_req = (ZhttpRequest *)sender();
 		RequestItem *ri = requests.value(_req);
 		assert(ri);
 
@@ -954,7 +983,7 @@ private slots:
 			if(close && !peerClosed)
 			{
 				peerClosed = true;
-				emit q->peerClosed();
+				q->peerClosed();
 				return;
 			}
 
@@ -962,9 +991,9 @@ private slots:
 			cleanup();
 
 			if(close)
-				emit q->closed();
+				q->closed();
 			else
-				emit q->error();
+				q->error();
 		}
 		else
 		{
@@ -980,7 +1009,7 @@ private slots:
 		}
 		else // WebSocketPassthrough
 		{
-			emit q->readyRead();
+			q->readyRead();
 		}
 	}
 
@@ -991,14 +1020,14 @@ private slots:
 
 	void sock_writeBytesChanged()
 	{
-		emit q->writeBytesChanged();
+		q->writeBytesChanged();
 	}
 
 	void sock_peerClosed()
 	{
 		peerCloseCode = sock->peerCloseCode();
 		peerCloseReason = sock->peerCloseReason();
-		emit q->peerClosed();
+		q->peerClosed();
 	}
 
 	void sock_closed()
@@ -1007,7 +1036,7 @@ private slots:
 		peerCloseReason = sock->peerCloseReason();
 		state = Idle;
 		cleanup();
-		emit q->closed();
+		q->closed();
 	}
 
 	void sock_error()
@@ -1015,9 +1044,10 @@ private slots:
 		state = Idle;
 		errorCondition = sock->errorCondition();
 		cleanup();
-		emit q->error();
+		q->error();
 	}
 
+private slots:
 	void doUpdate()
 	{
 		updating = false;
@@ -1026,7 +1056,7 @@ private slots:
 		{
 			state = Idle;
 			cleanup();
-			emit q->error();
+			q->error();
 			return;
 		}
 
@@ -1042,11 +1072,16 @@ private slots:
 				pendingWrittenFrames = 0;
 				pendingWrittenBytes = 0;
 
-				emit q->framesWritten(count, contentBytes);
+				q->framesWritten(count, contentBytes);
 			}
 		}
 	}
 
+	void doClosed()
+	{
+		q->closed();
+	}
+
 	void keepAliveTimer_timeout()
 	{
 		assert(mode != WebSocketPassthrough);
@@ -1066,7 +1101,7 @@ private slots:
 				// timeout while unconnected
 				state = Idle;
 				cleanup();
-				emit q->error();
+				q->error();
 			}
 		}
 		else
