@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2014-2022 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -26,7 +26,6 @@
 #include <assert.h>
 #include <QTimer>
 #include <QPointer>
-#include <QCoreApplication>
 #include <QRandomGenerator>
 #include "log.h"
 #include "bufferlist.h"
@@ -69,7 +68,13 @@ class WebSocketOverHttp::DisconnectManager : public QObject
 {
 	Q_OBJECT
 
-	map<WebSocketOverHttp *, Connection> disconnectedConnection;
+	struct WSConnections {
+		Connection disconnectedConnection;
+		Connection closedConnection;
+		Connection errorConnection;
+	};
+
+	map<WebSocketOverHttp *, WSConnections> wsConnectionMap;
 
 public:
 	DisconnectManager(QObject *parent = 0) :
@@ -80,9 +85,11 @@ public:
 	void addSocket(WebSocketOverHttp *sock)
 	{
 		sock->setParent(this);
-		disconnectedConnection[sock] = sock->disconnected.connect(boost::bind(&DisconnectManager::sock_disconnected, this, sock));
-		connect(sock, &WebSocketOverHttp::closed, this, &DisconnectManager::sock_closed);
-		connect(sock, &WebSocketOverHttp::error, this, &DisconnectManager::sock_error);
+		wsConnectionMap[sock] = { 
+			sock->disconnected.connect(boost::bind(&DisconnectManager::sock_disconnected, this, sock)),
+			sock->closed.connect(boost::bind(&DisconnectManager::sock_closed, this, sock)),
+			sock->error.connect(boost::bind(&DisconnectManager::sock_error, this, sock))
+		};
 
 		sock->sendDisconnect();
 	}
@@ -95,7 +102,7 @@ public:
 private:
 	void cleanupSocket(WebSocketOverHttp *sock)
 	{
-		disconnectedConnection.erase(sock);
+		wsConnectionMap.erase(sock);
 		delete sock;
 	}
 
@@ -105,22 +112,19 @@ private:
 		cleanupSocket(sock);
 	}
 
-private slots:
-	void sock_closed()
+	void sock_closed(WebSocketOverHttp *sock)
 	{
-		WebSocketOverHttp *sock = (WebSocketOverHttp *)sender();
 		cleanupSocket(sock);
 	}
 
-	void sock_error()
+	void sock_error(WebSocketOverHttp *sock)
 	{
-		WebSocketOverHttp *sock = (WebSocketOverHttp *)sender();
 		cleanupSocket(sock);
 	}
 };
 
-WebSocketOverHttp::DisconnectManager *WebSocketOverHttp::g_disconnectManager = 0;
-int WebSocketOverHttp::g_maxManagedDisconnects = -1;
+thread_local WebSocketOverHttp::DisconnectManager *WebSocketOverHttp::g_disconnectManager = 0;
+thread_local int WebSocketOverHttp::g_maxManagedDisconnects = -1;
 
 static QList<WsEvent> decodeEvents(const QByteArray &in, bool *ok = 0)
 {
@@ -189,6 +193,12 @@ class WebSocketOverHttp::Private : public QObject
 	Q_OBJECT
 
 public:
+	struct ReqConnections {
+		Connection readyReadConnection;
+		Connection bytesWrittenConnection;
+		Connection errorConnection;
+	};
+
 	WebSocketOverHttp *q;
 	ZhttpManager *zhttpManager;
 	QString connectHost;
@@ -227,9 +237,7 @@ public:
 	QTimer *retryTimer;
 	int retries;
 	int maxEvents;
-	Connection readyReadConnection;
-	Connection bytesWrittenConnection;
-	Connection errorConnection;
+	ReqConnections reqConnections;
 
 	Private(WebSocketOverHttp *_q) :
 		QObject(_q),
@@ -259,7 +267,7 @@ public:
 		maxEvents(0)
 	{
 		if(!g_disconnectManager)
-			g_disconnectManager = new DisconnectManager(QCoreApplication::instance());
+			g_disconnectManager = new DisconnectManager;
 
 		keepAliveTimer = new QTimer(this);
 		connect(keepAliveTimer, &QTimer::timeout, this, &Private::keepAliveTimer_timeout);
@@ -290,6 +298,7 @@ public:
 		disconnecting = false;
 		updateQueued = false;
 
+		reqConnections = ReqConnections();
 		delete req;
 		req = 0;
 
@@ -611,9 +620,11 @@ private:
 
 		req = zhttpManager->createRequest();
 		req->setParent(this);
-		readyReadConnection = req->readyRead.connect(boost::bind(&Private::req_readyRead, this));
-		bytesWrittenConnection = req->bytesWritten.connect(boost::bind(&Private::req_bytesWritten, this, boost::placeholders::_1));
-		errorConnection = req->error.connect(boost::bind(&Private::req_error, this));
+		reqConnections = {
+			req->readyRead.connect(boost::bind(&Private::req_readyRead, this)),
+			req->bytesWritten.connect(boost::bind(&Private::req_bytesWritten, this, boost::placeholders::_1)),
+			req->error.connect(boost::bind(&Private::req_error, this))
+		};
 
 		if(!connectHost.isEmpty())
 			req->setConnectHost(connectHost);
@@ -641,13 +652,12 @@ private:
 		req->endBody();
 	}
 
-private slots:
 	void req_readyRead()
 	{
 		if(inBuf.size() + req->bytesAvailable() > RESPONSE_BODY_MAX)
 		{
 			cleanup();
-			emit q->error();
+			q->error();
 			return;
 		}
 
@@ -667,6 +677,7 @@ private slots:
 		HttpHeaders responseHeaders = req->responseHeaders();
 		QByteArray responseBody = inBuf.take();
 
+		reqConnections = ReqConnections();
 		delete req;
 		req = 0;
 
@@ -691,7 +702,7 @@ private slots:
 				errorCondition = ErrorGeneric;
 
 			cleanup();
-			emit q->error();
+			q->error();
 			return;
 		}
 
@@ -727,7 +738,7 @@ private slots:
 		if(!ok)
 		{
 			cleanup();
-			emit q->error();
+			q->error();
 			return;
 		}
 
@@ -737,7 +748,7 @@ private slots:
 			if(events.isEmpty() && keepAliveInterval == -1)
 			{
 				cleanup();
-				emit q->error();
+				q->error();
 				return;
 			}
 
@@ -745,7 +756,7 @@ private slots:
 			if(!events.isEmpty() && events.first().type != "OPEN")
 			{
 				cleanup();
-				emit q->error();
+				q->error();
 				return;
 			}
 
@@ -838,21 +849,21 @@ private slots:
 
 		if(emitConnected)
 		{
-			emit q->connected();
+			q->connected();
 			if(!self)
 				return;
 		}
 
 		if(emitReadyRead)
 		{
-			emit q->readyRead();
+			q->readyRead();
 			if(!self)
 				return;
 		}
 
 		if(reqFrames > 0)
 		{
-			emit q->framesWritten(reqFrames, reqContentSize);
+			q->framesWritten(reqFrames, reqContentSize);
 			if(!self)
 				return;
 		}
@@ -864,7 +875,7 @@ private slots:
 
 		if(hadContent)
 		{
-			emit q->writeBytesChanged();
+			q->writeBytesChanged();
 			if(!self)
 				return;
 		}
@@ -877,12 +888,12 @@ private slots:
 			if(closeSent)
 			{
 				cleanup();
-				emit q->closed();
+				q->closed();
 				return;
 			}
 			else
 			{
-				emit q->peerClosed();
+				q->peerClosed();
 			}
 		}
 		else if(closeSent && keepAliveInterval == -1)
@@ -896,14 +907,14 @@ private slots:
 		if(disconnected)
 		{
 			cleanup();
-			emit q->error();
+			q->error();
 			return;
 		}
 
 		if(reqClose && peerClosing)
 		{
 			cleanup();
-			emit q->closed();
+			q->closed();
 			return;
 		}
 
@@ -947,6 +958,7 @@ private slots:
 				break;
 		}
 
+		reqConnections = ReqConnections();
 		delete req;
 		req = 0;
 
@@ -978,9 +990,10 @@ private slots:
 			errorCondition = WebSocket::ErrorTls;
 
 		cleanup();
-		emit q->error();
+		q->error();
 	}
 
+private slots:
 	void keepAliveTimer_timeout()
 	{
 		update();
@@ -996,7 +1009,7 @@ private slots:
 		cleanup();
 		errorCondition = pendingErrorCondition;
 		pendingErrorCondition = (ErrorCondition)-1;
-		emit q->error();
+		q->error();
 	}
 };
 
