@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2014 Fanout, Inc.
+ * Copyright (C) 2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -23,6 +24,7 @@
 #ifndef WSCONTROLMANAGER_H
 #define WSCONTROLMANAGER_H
 
+#include <memory>
 #include <QObject>
 #include "packet/wscontrolpacket.h"
 
@@ -33,25 +35,26 @@ class WsControlManager : public QObject
 	Q_OBJECT
 
 public:
-	WsControlManager(QObject *parent = 0);
+	WsControlManager();
 	~WsControlManager();
 
+	void setIdentity(const QByteArray &id);
 	void setIpcFileMode(int mode);
 
-	bool setInSpec(const QString &spec);
-	bool setOutSpec(const QString &spec);
+	bool setInitSpecs(const QStringList &specs);
+	bool setStreamSpecs(const QStringList &specs);
 
 	WsControlSession *createSession(const QByteArray &cid);
 
 private:
 	class Private;
-	Private *d;
+	std::unique_ptr<Private> d;
 
 	friend class WsControlSession;
 	void link(WsControlSession *s, const QByteArray &cid);
 	void unlink(const QByteArray &cid);
-	bool canWriteImmediately() const;
-	void write(const WsControlPacket::Item &item);
+	void writeInit(const WsControlPacket::Item &item);
+	void writeStream(const WsControlPacket::Item &item, const QByteArray &instanceAddress);
 	void registerKeepAlive(WsControlSession *s);
 	void unregisterKeepAlive(WsControlSession *s);
 };
