@@ -41,6 +41,13 @@
 #include "pushpinhandlerservice.h"
 #include "config.h"
 
+struct ServiceConnections{
+	Connection startedConnection;
+	Connection stoppedConnection;
+	Connection logConnection;
+	Connection errConnection;
+};
+
 static void trimlist(QStringList *list)
 {
 	for(int n = 0; n < list->count(); ++n)
@@ -246,10 +253,8 @@ static CommandLineParseResult parseCommandLine(QCommandLineParser *parser, ArgsD
 	return CommandLineOk;
 }
 
-class RunnerApp::Private : public QObject
+class RunnerApp::Private
 {
-	Q_OBJECT
-
 public:
 	RunnerApp *q;
 	ArgsData args;
@@ -258,9 +263,9 @@ public:
 	bool errored;
 	Connection quitConnection;
 	Connection hupConnection;
+	map<Service*, ServiceConnections> serviceConnectionMap;
 
 	Private(RunnerApp *_q) :
-		QObject(_q),
 		q(_q),
 		stopping(false),
 		errored(false)
@@ -584,7 +589,7 @@ public:
 			if(!serviceNames.contains("zurl") && CondureService::hasClientMode(condureBin))
 				useClient = true;
 
-			services += new CondureService("condure", condureBin, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("condure", defaultLevel), certsDir, clientBufferSize, clientMaxConnections, allowCompression, ports, useClient, this);
+			services += new CondureService("condure", condureBin, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("condure", defaultLevel), certsDir, clientBufferSize, clientMaxConnections, allowCompression, ports, useClient);
 		}
 
 		if(serviceNames.contains("mongrel2"))
@@ -605,7 +610,7 @@ public:
 			}
 
 			foreach(const ListenPort &p, ports)
-				services += new Mongrel2Service(m2Bin, QDir(runDir).filePath(QString("%1mongrel2.sqlite").arg(filePrefix)), "default_" + QString::number(p.port), runDir, !args.mergeOutput ? logDir : QString(), filePrefix, p.port, p.ssl, logLevels.value("mongrel2", defaultLevel), this);
+				services += new Mongrel2Service(m2Bin, QDir(runDir).filePath(QString("%1mongrel2.sqlite").arg(filePrefix)), "default_" + QString::number(p.port), runDir, !args.mergeOutput ? logDir : QString(), filePrefix, p.port, p.ssl, logLevels.value("mongrel2", defaultLevel));
 		}
 
 		if(serviceNames.contains("m2adapter"))
@@ -614,7 +619,7 @@ public:
 			foreach(const ListenPort &p, ports)
 				portsOnly += p.port;
 
-			services += new M2AdapterService(m2aBin, QDir(libDir).filePath("m2adapter.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("m2adapter", defaultLevel), portsOnly, this);
+			services += new M2AdapterService(m2aBin, QDir(libDir).filePath("m2adapter.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("m2adapter", defaultLevel), portsOnly);
 		}
 
 		bool quietCheck = false;
@@ -625,26 +630,28 @@ public:
 			if(settings.contains("runner/zurl_bin"))
 				zurlBin = settings.value("runner/zurl_bin").toString();
 
-			services += new ZurlService(zurlBin, QDir(libDir).filePath("zurl.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("zurl", defaultLevel), this);
+			services += new ZurlService(zurlBin, QDir(libDir).filePath("zurl.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("zurl", defaultLevel));
 
 			// when zurl is managed by pushpin, log updates checks as debug level
 			quietCheck = true;
 		}
 
 		if(serviceNames.contains("pushpin-proxy"))
-			services += new PushpinProxyService(proxyBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("pushpin-proxy", defaultLevel), args.routeLines, quietCheck, this);
+			services += new PushpinProxyService(proxyBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevels.value("pushpin-proxy", defaultLevel), args.routeLines, quietCheck);
 
 		if(serviceNames.contains("pushpin-handler"))
-			services += new PushpinHandlerService(handlerBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, portOffset, logLevels.value("pushpin-handler", defaultLevel), this);
+			services += new PushpinHandlerService(handlerBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, portOffset, logLevels.value("pushpin-handler", defaultLevel));
 
 		foreach(Service *s, services)
 		{
-			connect(s, &Service::started, this, &Private::service_started);
-			connect(s, &Service::stopped, this, &Private::service_stopped);
-			connect(s, &Service::logLine, this, &Private::service_logLine);
-			connect(s, &Service::error, this, &Private::service_error);
+			serviceConnectionMap[s] = {
+				s->started.connect(boost::bind(&Private::service_started, this)),
+				s->stopped.connect(boost::bind(&Private::service_stopped, this, s)),
+				s->logLine.connect(boost::bind(&Private::service_logLine, this, boost::placeholders::_1, s)),
+				s->error.connect(boost::bind(&Private::service_error, this, boost::placeholders::_1, s))
+			};
 
-			if(!args.mergeOutput || qobject_cast<Mongrel2Service*>(s))
+			if(!args.mergeOutput || s->alwaysLogStatus())
 				log_info("starting %s", qPrintable(s->name()));
 
 			s->start();
@@ -680,7 +687,7 @@ private:
 	{
 		foreach(Service *s, services)
 		{
-			if(!args.mergeOutput || qobject_cast<Mongrel2Service*>(s))
+			if(!args.mergeOutput || s->alwaysLogStatus())
 				log_info("stopping %s", qPrintable(s->name()));
 
 			s->stop();
@@ -701,7 +708,6 @@ private:
 		q->quit(errored ? 1 : 0);
 	}
 
-private slots:
 	void service_started()
 	{
 		bool allStarted = true;
@@ -718,9 +724,9 @@ private slots:
 			log_info("started");
 	}
 
-	void service_stopped()
+	void service_stopped(Service *s)
 	{
-		Service *s = (Service *)sender();
+		serviceConnectionMap.erase(s);
 
 		services.removeAll(s);
 		delete s;
@@ -728,22 +734,20 @@ private slots:
 		checkStopped();
 	}
 
-	void service_logLine(const QString &line)
+	void service_logLine(const QString &line, Service *s)
 	{
-		Service *s = (Service *)sender();
-
 		QString out = tryInsertPrefix(s->formatLogLine(line), '[' + s->name() + "] ");
 		if(!out.isEmpty()) {
 			log_raw(qPrintable(out));
 		}
 	}
 
-	void service_error(const QString &error)
+	void service_error(const QString &error, Service *s)
 	{
-		Service *s = (Service *)sender();
-
 		log_error("%s: %s", qPrintable(s->name()), qPrintable(error));
 
+		serviceConnectionMap.erase(s);
+
 		services.removeAll(s);
 		delete s;
 
@@ -761,7 +765,6 @@ private slots:
 		}
 	}
 
-private:
 	void reload()
 	{
 		log_info("reloading");
@@ -796,20 +799,14 @@ private:
 	}
 };
 
-RunnerApp::RunnerApp(QObject *parent) :
-	QObject(parent)
-{
-	d = new Private(this);
+RunnerApp::RunnerApp() {
+	d = std::make_unique<Private>(this);
 }
 
-RunnerApp::~RunnerApp()
-{
-	delete d;
-}
+RunnerApp::~RunnerApp() = default;
 
 void RunnerApp::start()
 {
 	d->start();
 }
 
-#include "runnerapp.moc"
