@@ -23,18 +23,17 @@
 #ifndef RUNNERAPP_H
 #define RUNNERAPP_H
 
-#include <QObject>
 #include <boost/signals2.hpp>
+#include <map>
 
+using std::map;
 using SignalInt = boost::signals2::signal<void(int)>;
 using Connection = boost::signals2::scoped_connection;
 
-class RunnerApp : public QObject
+class RunnerApp
 {
-	Q_OBJECT
-
 public:
-	RunnerApp(QObject *parent = 0);
+	RunnerApp();
 	~RunnerApp();
 
 	void start();
@@ -44,7 +43,7 @@ public:
 private:
 	class Private;
 	friend class Private;
-	Private *d;
+	std::unique_ptr<Private> d;
 };
 
 #endif
