@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2014-2023 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -44,6 +44,7 @@
 #define ACTIVITY_TIMEOUT 100
 #define REFRESH_INTERVAL 1000
 #define EXTERNAL_CONNECTIONS_MAX_INTERVAL 10000
+#define EXPIRE_MAX 10000
 
 #define SHOULD_PROCESS_TIME(x) (x * 3 / 4)
 
@@ -96,7 +97,7 @@ public:
 		bool linger;
 		qint64 lastReport;
 		qint64 retrySeq;
-		QByteArray from; // external
+		QByteArray from; // external or linger source
 		int ttl; // external
 		qint64 lastActive; // external
 
@@ -207,6 +208,18 @@ public:
 		}
 	};
 
+	class RetryInfo
+	{
+	public:
+		quint64 nextSeq;
+		QMap<quint64, ConnectionInfo*> connectionInfoBySeq;
+
+		RetryInfo() :
+			nextSeq(0)
+		{
+		}
+	};
+
 	class Report
 	{
 	public:
@@ -397,7 +410,7 @@ public:
 	QHash<QByteArray, quint32> routeActivity;
 	QHash<QByteArray, ConnectionInfo*> connectionInfoById;
 	QHash<QByteArray, QSet<ConnectionInfo*> > connectionInfoByRoute;
-	QMap<quint64, ConnectionInfo*> connectionInfoByRetrySeq;
+	QHash<QByteArray, RetryInfo> retryInfoBySource;
 	QVector<QSet<ConnectionInfo*> > connectionInfoRefreshBuckets;
 	int currentConnectionInfoRefreshBucket;
 	QHash<QByteArray, QHash<QByteArray, ConnectionInfo*> > externalConnectionInfoByFrom;
@@ -412,7 +425,6 @@ public:
 	QHash<QByteArray, Report*> reports;
 	Counts combinedCounts;
 	Report combinedReport;
-	quint64 nextRetrySeq;
 	QTimer *activityTimer;
 	QTimer *reportTimer;
 	QTimer *refreshTimer;
@@ -439,7 +451,6 @@ public:
 		currentConnectionInfoRefreshBucket(0),
 		currentSubscriptionRefreshBucket(0),
 		wheel(TimerWheel((_connectionsMax * 2) + _subscriptionsMax)),
-		nextRetrySeq(0),
 		reportTimer(0)
 	{
 		activityTimer = new QTimer(this);
@@ -472,28 +483,36 @@ public:
 	{
 		if(activityTimer)
 		{
-			activityTimer->setParent(0);
 			activityTimer->disconnect(this);
+			activityTimer->setParent(0);
 			activityTimer->deleteLater();
 			activityTimer = 0;
 		}
 
 		if(reportTimer)
 		{
-			reportTimer->setParent(0);
 			reportTimer->disconnect(this);
+			reportTimer->setParent(0);
 			reportTimer->deleteLater();
 			reportTimer = 0;
 		}
 
 		if(refreshTimer)
 		{
-			refreshTimer->setParent(0);
 			refreshTimer->disconnect(this);
+			refreshTimer->setParent(0);
 			refreshTimer->deleteLater();
 			refreshTimer = 0;
 		}
 
+		if(externalConnectionsMaxTimer)
+		{
+			externalConnectionsMaxTimer->disconnect(this);
+			externalConnectionsMaxTimer->setParent(0);
+			externalConnectionsMaxTimer->deleteLater();
+			externalConnectionsMaxTimer = 0;
+		}
+
 		qDeleteAll(connectionInfoById);
 
 		QMutableHashIterator<QByteArray, QHash<QByteArray, ConnectionInfo*> > it(externalConnectionInfoByFrom);
@@ -537,6 +556,7 @@ public:
 		{
 			if(!prometheusServer->listenLocal(portStr.mid(6)))
 			{
+				promServerConnection.disconnect();
 				delete prometheusServer;
 
 				return false;
@@ -560,6 +580,7 @@ public:
 
 			if(!prometheusServer->listen(addr, port))
 			{
+				promServerConnection.disconnect();
 				delete prometheusServer;
 
 				return false;
@@ -621,8 +642,8 @@ public:
 		}
 		else if(reportInterval <= 0 && reportTimer)
 		{
-			reportTimer->setParent(0);
 			reportTimer->disconnect(this);
+			reportTimer->setParent(0);
 			reportTimer->deleteLater();
 			reportTimer = 0;
 		}
@@ -718,7 +739,15 @@ public:
 		}
 
 		if(c->retrySeq >= 0)
-			connectionInfoByRetrySeq.remove(c->retrySeq);
+		{
+			RetryInfo &ri = retryInfoBySource[c->from];
+			ri.connectionInfoBySeq.remove(c->retrySeq);
+
+			// FIXME: we keep the source entry even when there are no more
+			// connections, to avoid resetting the seq value. if there is
+			// a lot of proxy instance churn, retryInfoBySource could
+			// fill up with unused entries that will never be cleaned up.
+		}
 
 		if(c->lastRefresh >= 0)
 		{
@@ -764,20 +793,25 @@ public:
 		wheelRemove(c);
 	}
 
-	void removeLingeringConnections(quint64 retrySeq)
+	void removeLingeringConnections(const QByteArray &source, quint64 retrySeq)
 	{
+		if(!retryInfoBySource.contains(source))
+			return;
+
+		RetryInfo &ri = retryInfoBySource[source];
+
 		// invalid retry seq
-		if(retrySeq >= nextRetrySeq)
+		if(retrySeq >= ri.nextSeq)
 			return;
 
 		QList<ConnectionInfo*> toRemove;
 
-		QMap<quint64, ConnectionInfo*>::iterator it = connectionInfoByRetrySeq.find(retrySeq);
-		while(it != connectionInfoByRetrySeq.end())
+		QMap<quint64, ConnectionInfo*>::iterator it = ri.connectionInfoBySeq.find(retrySeq);
+		while(it != ri.connectionInfoBySeq.end())
 		{
 			toRemove += it.value();
 
-			if(it == connectionInfoByRetrySeq.begin())
+			if(it == ri.connectionInfoBySeq.begin())
 				break;
 
 			--it;
@@ -992,7 +1026,7 @@ public:
 
 	void sendConnectionsMax(const QByteArray &routeId, ConnectionsMax *cm, qint64 now)
 	{
-		emit q->connMax(getConnMaxPacket(routeId, cm, now));
+		q->connMax(getConnMaxPacket(routeId, cm, now));
 	}
 
 	void updateConnectionsMax(const QByteArray &routeId, qint64 now)
@@ -1074,7 +1108,7 @@ public:
 		QList<QByteArray> refreshedConnIds;
 		QSet<QByteArray> routesUpdated;
 
-		while(true)
+		for(int i = 0; i < EXPIRE_MAX; ++i)
 		{
 			TimerWheel::Expired expired = wheel.takeExpired();
 
@@ -1151,7 +1185,7 @@ public:
 						removeSubscription(s);
 						delete s;
 
-						emit q->unsubscribed(mode, channel);
+						q->unsubscribed(mode, channel);
 					}
 					else
 					{
@@ -1167,7 +1201,7 @@ public:
 		}
 
 		if(!refreshedConnIds.isEmpty())
-			emit q->connectionsRefreshed(refreshedConnIds);
+			q->connectionsRefreshed(refreshedConnIds);
 
 		foreach(const QByteArray &routeId, routesUpdated)
 			updateConnectionsMax(routeId, now);
@@ -1195,7 +1229,7 @@ public:
 		}
 
 		if(!refreshedIds.isEmpty())
-			emit q->connectionsRefreshed(refreshedIds);
+			q->connectionsRefreshed(refreshedIds);
 
 		++currentConnectionInfoRefreshBucket;
 		if(currentConnectionInfoRefreshBucket >= connectionInfoRefreshBuckets.count())
@@ -1280,7 +1314,7 @@ public:
 	void mergeExternalConnectionsMax(const StatsPacket &packet, qint64 now)
 	{
 		if(packet.retrySeq >= 0)
-			removeLingeringConnections((quint64)packet.retrySeq);
+			removeLingeringConnections(packet.from, (quint64)packet.retrySeq);
 
 		QHash<QByteArray, ExternalConnectionsMax> &maxes = externalConnectionsMaxes[packet.route].maxes;
 
@@ -1396,7 +1430,7 @@ public:
 		if(sock)
 			write(p);
 
-		emit q->reported(QList<StatsPacket>() << p);
+		q->reported(QList<StatsPacket>() << p);
 	}
 
 	StatsPacket getConnMaxPacket(const QByteArray &routeId, ConnectionsMax *cm, qint64 now)
@@ -1499,7 +1533,7 @@ private slots:
 		}
 
 		if(!reportPackets.isEmpty())
-			emit q->reported(reportPackets);
+			q->reported(reportPackets);
 	}
 
 	void refresh_timeout()
@@ -1747,7 +1781,7 @@ void StatsManager::addConnection(const QByteArray &id, const QByteArray &routeId
 		d->sendConnected(c);
 }
 
-int StatsManager::removeConnection(const QByteArray &id, bool linger)
+int StatsManager::removeConnection(const QByteArray &id, bool linger, const QByteArray &source)
 {
 	Private::ConnectionInfo *c = d->connectionInfoById.value(id);
 	if(!c)
@@ -1765,9 +1799,16 @@ int StatsManager::removeConnection(const QByteArray &id, bool linger)
 		if(!c->linger)
 		{
 			c->linger = true;
-			c->retrySeq = (qint64)d->nextRetrySeq++;
 
-			d->connectionInfoByRetrySeq.insert((quint64)c->retrySeq, c);
+			if(!d->retryInfoBySource.contains(source))
+				d->retryInfoBySource.insert(source, Private::RetryInfo());
+
+			Private::RetryInfo &ri = d->retryInfoBySource[source];
+
+			c->from = source;
+			c->retrySeq = (qint64)ri.nextSeq++;
+
+			ri.connectionInfoBySeq.insert((quint64)c->retrySeq, c);
 
 			// hack to ensure full linger time honored by refresh processing
 			qint64 lingerStartTime = now + (d->connectionLinger - SHOULD_PROCESS_TIME(d->connectionTtl));
@@ -1876,7 +1917,7 @@ void StatsManager::removeSubscription(const QString &mode, const QString &channe
 		d->removeSubscription(s);
 		delete s;
 
-		emit unsubscribed(mode, channel);
+		unsubscribed(mode, channel);
 	}
 }
 
@@ -2087,9 +2128,14 @@ void StatsManager::flushReport(const QByteArray &routeId)
 	d->flushReport(routeId);
 }
 
-qint64 StatsManager::lastRetrySeq() const
+qint64 StatsManager::lastRetrySeq(const QByteArray &source) const
 {
-	return ((qint64)d->nextRetrySeq) - 1;
+	if(!d->retryInfoBySource.contains(source))
+		return -1;
+
+	Private::RetryInfo &ri = d->retryInfoBySource[source];
+
+	return ((qint64)ri.nextSeq) - 1;
 }
 
 StatsPacket StatsManager::getConnMaxPacket(const QByteArray &routeId)
