@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2013-2022 Fanout, Inc.
+ * Copyright (C) 2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -21,8 +22,8 @@
  */
 
 #include <unistd.h>
+#include <boost/signals2.hpp>
 #include <QtTest/QtTest>
-#include <QSignalSpy>
 #include <QDir>
 #include <QJsonDocument>
 #include <QJsonObject>
@@ -39,6 +40,7 @@
 #include "rtimer.h"
 #include "zhttpmanager.h"
 #include "statsmanager.h"
+#include "domainmap.h"
 #include "engine.h"
 
 Q_DECLARE_METATYPE(QList<StatsPacket>);
@@ -85,6 +87,11 @@ public:
 	int clientReqsFinished;
 	QByteArray requestBody;
 	QHash<QByteArray, HttpResponseData> responses;
+	Connection zhttpClientInValveConnection;
+	Connection zhttpServerInValveConnection;
+	Connection zhttpServerInStreamValveConnection;
+	Connection handlerAcceptValveConnection;
+	Connection handlerInspectValveConnection;
 
 	Wrapper(QObject *parent, QDir _workDir) :
 		QObject(parent),
@@ -106,16 +113,16 @@ public:
 
 		zhttpClientInSock = new QZmq::Socket(QZmq::Socket::Sub, this);
 		zhttpClientInValve = new QZmq::Valve(zhttpClientInSock, this);
-		connect(zhttpClientInValve, &QZmq::Valve::readyRead, this, &Wrapper::zhttpClientIn_readyRead);
+		zhttpClientInValveConnection = zhttpClientInValve->readyRead.connect(boost::bind(&Wrapper::zhttpClientIn_readyRead, this, boost::placeholders::_1));
 
 		zhttpServerInSock = new QZmq::Socket(QZmq::Socket::Pull, this);
 		zhttpServerInValve = new QZmq::Valve(zhttpServerInSock, this);
-		connect(zhttpServerInValve, &QZmq::Valve::readyRead, this, &Wrapper::zhttpServerIn_readyRead);
+		zhttpServerInValveConnection = zhttpServerInValve->readyRead.connect(boost::bind(&Wrapper::zhttpServerIn_readyRead, this, boost::placeholders::_1));
 
 		zhttpServerInStreamSock = new QZmq::Socket(QZmq::Socket::Router, this);
 		zhttpServerInStreamSock->setIdentity("test-server");
 		zhttpServerInStreamValve = new QZmq::Valve(zhttpServerInStreamSock, this);
-		connect(zhttpServerInStreamValve, &QZmq::Valve::readyRead, this, &Wrapper::zhttpServerInStream_readyRead);
+		zhttpServerInStreamValveConnection = zhttpServerInStreamValve->readyRead.connect(boost::bind(&Wrapper::zhttpServerInStream_readyRead, this, boost::placeholders::_1));
 
 		zhttpServerOutSock = new QZmq::Socket(QZmq::Socket::Pub, this);
 
@@ -125,12 +132,12 @@ public:
 
 		handlerAcceptSock = new QZmq::Socket(QZmq::Socket::Router, this);
 		handlerAcceptValve = new QZmq::Valve(handlerAcceptSock, this);
-		connect(handlerAcceptValve, &QZmq::Valve::readyRead, this, &Wrapper::handlerAccept_readyRead);
+		handlerAcceptValveConnection = handlerAcceptValve->readyRead.connect(boost::bind(&Wrapper::handlerAccept_readyRead, this, boost::placeholders::_1));
 
 		handlerInspectValve = new QZmq::Valve(handlerInspectSock, this);
-		connect(handlerInspectValve, &QZmq::Valve::readyRead, this, &Wrapper::handlerInspect_readyRead);
+		handlerInspectValveConnection = handlerInspectValve->readyRead.connect(boost::bind(&Wrapper::handlerInspect_readyRead, this, boost::placeholders::_1));
 
-		handlerRetryOutSock = new QZmq::Socket(QZmq::Socket::Push, this);
+		handlerRetryOutSock = new QZmq::Socket(QZmq::Socket::Router, this);
 	}
 
 	void startHttp()
@@ -178,7 +185,7 @@ public:
 		responses.clear();
 	}
 
-private slots:
+private:
 	void zhttpClientIn_readyRead(const QList<QByteArray> &message)
 	{
 		log_debug("client in");
@@ -535,7 +542,12 @@ private slots:
 				vretry["request-data"] = vaccept["request-data"];
 				QByteArray buf = TnetString::fromVariant(vretry);
 				log_debug("retrying: %s", qPrintable(TnetString::variantToString(vretry, -1)));
-				handlerRetryOutSock->write(QList<QByteArray>() << buf);
+
+				QList<QByteArray> msg;
+				msg.append("proxy");
+				msg.append(QByteArray());
+				msg.append(buf);
+				handlerRetryOutSock->write(msg);
 				return;
 			}
 		}
@@ -552,14 +564,22 @@ class ProxyEngineTest : public QObject
 	Q_OBJECT
 
 private:
+	DomainMap *domainMap;
 	Engine *engine;
 	Wrapper *wrapper;
+	QList<StatsPacket> trackedPackets;
 
 private:
 	void reset()
 	{
 		wrapper->reset();
 		engine->statsManager()->flushReport(QByteArray());
+		trackedPackets.clear();		
+	}
+	
+	void appendTrackedPackets(const QList<StatsPacket>& packets)
+	{
+		trackedPackets.append(packets);
 	}
 
 private slots:
@@ -578,7 +598,9 @@ private slots:
 		wrapper = new Wrapper(this, workDir);
 		wrapper->startHttp();
 
-		engine = new Engine(this);
+		domainMap = new DomainMap(configDir.filePath("routes"), this);
+
+		engine = new Engine(domainMap, this);
 
 		Engine::Configuration config;
 		config.clientId = "proxy";
@@ -592,12 +614,11 @@ private slots:
 		config.acceptSpec = ("ipc://" + workDir.filePath("accept"));
 		config.retryInSpec = ("ipc://" + workDir.filePath("retry-out"));
 		config.statsSpec = ("ipc://" + workDir.filePath("stats"));
+		config.sessionsMax = 20;
 		config.inspectTimeout = 500;
 		config.inspectPrefetch = 5;
-		config.routesFile = configDir.filePath("routes");
 		config.sigIss = "pushpin";
 		config.sigKey = Jwt::EncodingKey::fromSecret("changeme");
-		config.connectionsMax = 20;
 		config.statsConnectionTtl = 120;
 		config.statsReportInterval = 1000; // set a large interval so there's only one working report
 		QVERIFY(engine->start(config));
@@ -610,8 +631,10 @@ private slots:
 	void cleanupTestCase()
 	{
 		delete engine;
+		delete domainMap;
 		delete wrapper;
 
+		QCoreApplication::instance()->sendPostedEvents();
 		RTimer::deinit();
 	}
 
@@ -619,7 +642,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(	
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -643,12 +668,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 23); // "GET" + "/path?a=b" + "Host" + "example"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 43); // "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11"
@@ -740,7 +762,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+            		boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -785,12 +809,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 9); // "POST" + "/path"
 		QCOMPARE(p.clientContentBytesReceived, 11); // "hello world"
 		QCOMPARE(p.clientHeaderBytesSent, 43); // "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11"
@@ -805,7 +826,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -847,12 +870,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 9); // "POST" + "/path"
 		QCOMPARE(p.clientContentBytesReceived, 5); // "hello"
 		QCOMPARE(p.clientHeaderBytesSent, 0);
@@ -867,7 +887,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -891,12 +913,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 22); // "GET" + "/path?hold=response"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 0);
@@ -911,7 +930,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -936,12 +957,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 20); // "GET" + "/path?hold=stream"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 0);
@@ -977,7 +995,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -999,12 +1019,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 18); // "GET" + "/path?hold=none"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 43); // "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11"
@@ -1040,7 +1057,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -1066,12 +1085,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 31); // "GET" + "/path?hold=stream&large=true"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 5); // "200" + "OK"
@@ -1086,7 +1102,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -1111,12 +1129,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 29); // "GET" + "/path?hold=none&large=true"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 27); // "200" + "OK" + "Content-Type" + "text/plain"
@@ -1131,7 +1146,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -1166,12 +1183,9 @@ private slots:
 		headerBytes += ZhttpManager::estimateRequestHeaderBytes(req2Data.method, req2Data.uri, req2Data.headers);
 		contentBytes += req2Data.body.size();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 42); // "GET" + "/path2?hold=response&body-instruct=true"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 43); // "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11"
@@ -1186,7 +1200,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		wrapper->sharingKey = "test";
 
@@ -1228,12 +1244,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 16); // "GET" + "/path" + "GET" + "/path"
 		QCOMPARE(p.clientContentBytesReceived, 0);
 		QCOMPARE(p.clientHeaderBytesSent, 86); // "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11" + "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11"
@@ -1248,7 +1261,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		wrapper->sharingKey = "test";
 
@@ -1321,12 +1336,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 18); // "POST" + "/path" + "POST" + "/path"
 		QCOMPARE(p.clientContentBytesReceived, 22); // "hello world" + "hello world"
 		QCOMPARE(p.clientHeaderBytesSent, 86); // "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11" + "200" + "OK" + "Content-Type" + "text/plain" + "Content-Length" + "11"
@@ -1341,7 +1353,9 @@ private slots:
 	{
 		reset();
 
-		QSignalSpy spy(engine->statsManager(), SIGNAL(reported(const QList<StatsPacket> &)));
+		boost::signals2::scoped_connection reportConnection = engine->statsManager()->reported.connect(
+			boost::bind(&ProxyEngineTest::appendTrackedPackets, this, boost::placeholders::_1)
+		);
 
 		ZhttpRequestPacket zreq;
 		zreq.from = "test-client";
@@ -1379,12 +1393,9 @@ private slots:
 
 		HttpRequestData reqData = QHashIterator<QByteArray, HttpRequestData>(wrapper->serverReqs).next().value();
 
-		QCOMPARE(spy.count(), 1);
+		QCOMPARE(trackedPackets.size(), 1);
 
-		QList<StatsPacket> packets = qvariant_cast<QList<StatsPacket>>(spy.takeFirst().at(0));
-		QCOMPARE(packets.count(), 1);
-
-		StatsPacket p = packets[0];
+		StatsPacket p = trackedPackets.takeFirst();
 		QCOMPARE(p.clientHeaderBytesReceived, 8); // "GET" + "/path"
 		QCOMPARE(p.clientContentBytesReceived, 5);
 		QCOMPARE(p.clientHeaderBytesSent, 22); // "101" + "Switching Protocols"
