@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2021-2022 Fanout, Inc.
- * Copyright (C) 2023 Fastly, Inc.
+ * Copyright (C) 2023-2024 Fastly, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -32,6 +32,9 @@ use std::os::raw::c_char;
 use std::ptr;
 use std::slice;
 
+#[cfg(test)]
+use crate::import_cpptest;
+
 #[repr(C)]
 pub struct ExpiredTimer {
     key: libc::c_int,
@@ -380,6 +383,7 @@ const WZMQ_TCP_KEEPALIVE: libc::c_int = 10;
 const WZMQ_TCP_KEEPALIVE_IDLE: libc::c_int = 11;
 const WZMQ_TCP_KEEPALIVE_CNT: libc::c_int = 12;
 const WZMQ_TCP_KEEPALIVE_INTVL: libc::c_int = 13;
+const WZMQ_ROUTER_MANDATORY: libc::c_int = 14;
 
 // NOTE: must match values in wzmq.h
 const WZMQ_DONTWAIT: libc::c_int = 0x01;
@@ -722,6 +726,21 @@ pub unsafe extern "C" fn wzmq_setsockopt(
                 return -1;
             }
         }
+        WZMQ_ROUTER_MANDATORY => {
+            if option_len as u32 != libc::c_int::BITS / 8 {
+                return -1;
+            }
+
+            let x = match (option_value as *mut libc::c_int).as_ref() {
+                Some(x) => x,
+                None => return -1,
+            };
+
+            if let Err(e) = sock.set_router_mandatory(*x != 0) {
+                set_errno(e.to_raw());
+                return -1;
+            }
+        }
         WZMQ_SNDHWM => {
             if option_len as u32 != libc::c_int::BITS / 8 {
                 return -1;
@@ -1091,35 +1110,8 @@ pub unsafe fn build_config_destroy(c: *mut BuildConfig) {
     drop(CString::from_raw(c.lib_dir));
 }
 
-#[cfg(all(test, target_os = "macos"))]
-#[link(name = "pushpin-cpptest")]
-#[link(name = "pushpin-cpp")]
-#[link(name = "QtCore", kind = "framework")]
-#[link(name = "QtNetwork", kind = "framework")]
-#[link(name = "QtTest", kind = "framework")]
-#[link(name = "c++")]
-extern "C" {
-    pub fn httpheaders_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn jwt_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn routesfile_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn proxyengine_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn jsonpatch_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn instruct_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn idformat_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn publishformat_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn publishitem_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn handlerengine_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-    pub fn template_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
-}
-
-#[cfg(all(test, not(target_os = "macos")))]
-#[link(name = "pushpin-cpptest")]
-#[link(name = "pushpin-cpp")]
-#[link(name = "Qt5Core")]
-#[link(name = "Qt5Network")]
-#[link(name = "Qt5Test")]
-#[link(name = "stdc++")]
-extern "C" {
+#[cfg(test)]
+import_cpptest! {
     pub fn httpheaders_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
     pub fn jwt_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
     pub fn routesfile_test(argc: libc::c_int, argv: *const *const libc::c_char) -> libc::c_int;
